"""
Trial Name
===========

This utility method create a name for the RLlib trial experiment.
"""

from ray.tune.experiment.trial import Trial
from eprllib import logger
from eprllib.Utils.annotations import trial_str_creator_for_tune

def trial_str_creator(trial: Trial, name:str='eprllib'):
    """
    This method create a description for the folder where the outputs and checkpoints 
    will be save.

    Args:
        trial: A trial type of RLlib.
        name (str): Optional name for the trial. Default: eprllib

    Returns:
        str: Return a unique string for the folder of the trial.
    """
    logger.warning("TrialStrCreator: This method is deprecated. Use trial_str_creator_for_tune instead.")
        
    trial_str_creator_for_tune(trial, name)
    



"""
Reward Function
================

This module contains the base class for defining reward functions.

It is preferred to use the ``infos`` dictionary and not the observation, since the latter is 
a numpy array and cannot be called by key values, which is prone to errors when developing the program 
and indexing an array may change.

The terminated and truncated flags are arguments in the reward function ``get_reward`` method to allow
implementations with dispersed reward. This flags allow return the final reward when the episode ends.
"""
from typing import Dict, Any, List
from numpy.typing import NDArray
from numpy import floating

from eprllib import logger
from eprllib.Utils.annotations import OverrideToImplementCustomLogic

class BaseReward:
    """
    This class is the base class for defining reward functions.
    """
    reward_fn_config: Dict[str, Any] = {}
    agent_name: str = ""
    cumulated_reward: List[float] = []
    reward_timestep: int = 0
    
    def __init__(
        self,
        agent_name: str,
        reward_fn_config: Dict[str, Any] = {}
    ) -> None:
        """
        Initializes the base reward function with the given configuration.

        Args:
            reward_fn_config (Dict[str, Any]): Configuration dictionary for the reward function.
        """
        self.reward_fn_config = reward_fn_config
        self.agent_name = agent_name
        
        self.cumulated_reward = []
        self.reward_timestep = 0
        
        logger.info(f"BaseReward: The BaseReward was correctly inicializated with {self.reward_fn_config} config.")
        
        # Make sure, `setup()` is only called once, no matter what.
        if hasattr(self, "_is_setup") and self._is_setup:
            raise RuntimeError(
                "``BaseActionMapper.setup()`` called twice within your ActionMapper implementation "
                f"{self}! Make sure you are using the proper inheritance order "
                " and that you are NOT overriding the constructor, but "
                "only the ``setup()`` method of your subclass."
            )
        try:
            self.setup()
        except AttributeError as e:
            raise e

        self._is_setup:bool = True
        self._set_initial_parameters_once: bool = False
    
    
    def reset(self) -> None:
        """
        Resets the reward function to its initial state.
        """
        self.cumulated_reward = []
        self.reward_timestep = 0
    
    
    def set_initial_parameters(
        self,
        obs_indexed: Dict[str, int]
    ) -> None:
        if not self._set_initial_parameters_once:
            self._set_initial_parameters(obs_indexed)
            self._set_initial_parameters_once = True
            logger.info(f"BaseReward: The initial parameters were correctly set.")
        else:
            logger.info(f"BaseReward: The initial parameters were already set.")
    
    
    # ===========================
    # === OVERRIDABLE METHODS ===
    # ===========================
    
    @OverrideToImplementCustomLogic
    def setup(self) -> None:
        """
        This method can be overridden in subclasses to perform setup tasks.
        """
        pass
    
    @OverrideToImplementCustomLogic
    def _set_initial_parameters(
        self,
        obs_indexed: Dict[str, int]
    ) -> None:
        """
        This method can be overridden in subclasses to set initial parameters based on the provided infos.
        
        Example:
            ```
            self.zone_people_occupant_count_index = obs_indexed[get_variable_name(
                self.agent_name, 
                "Zone People Occupant Count", 
                self.reward_fn_config['thermal_zone']
            )]
            ```

        Args:
            infos (Dict[str, Any]): The infos dictionary containing necessary information for initialization.
        """
        pass
    
    @OverrideToImplementCustomLogic
    def get_reward(
        self,
        prev_obs: NDArray[floating[Any]],
        prev_action: Any,
        obs: NDArray[floating[Any]],
        terminated: bool,
        truncated: bool
        ) -> float:
        """
        This method must be implemented in the subclass to calculate the reward.

        Args:
            infos (Dict[str, Any]): The infos dictionary containing the necessary information for calculating the reward.
            terminated (bool): Indicates if the episode has terminated.
            truncated (bool): Indicates if the episode has been truncated.

        Returns:
            float: The calculated reward.
        """
        msg = "BaseReward: This method must be implemented in the subclass."
        logger.error(msg)
        raise NotImplementedError(msg)





import pytest

from eprllib.Agents import SIMULATION_PARAMETERS, ZONE_SIMULATION_PARAMETERS, PREDICTION_VARIABLES #, PREDICTION_HOURS
from eprllib.Agents.ObservationSpec import ObservationSpec

class TestObservationspec:

    def test___getitem___1(self):
        """
        Test that the __getitem__ method correctly retrieves attribute values.
        This test checks if the method returns the expected value for a given key.
        """
        obs_spec = ObservationSpec(variables=[('test_var', 'test_unit')])
        assert obs_spec['variables'] == [('test_var', 'test_unit')]

    def test___getitem___nonexistent_attribute(self):
        """
        Test that attempting to access a non-existent attribute raises an AttributeError.
        This tests the edge case of trying to access an attribute that doesn't exist in the ObservationSpec object.
        """
        obs_spec = ObservationSpec()
        with pytest.raises(AttributeError):
            obs_spec['nonexistent_attribute']

    def test___setitem___1(self):
        """
        Test that __setitem__ correctly sets an attribute of the ObservationSpec instance.
        """
        obs_spec = ObservationSpec()
        obs_spec['variables'] = [('var1', 'type1'), ('var2', 'type2')]
        assert obs_spec.variables == [('var1', 'type1'), ('var2', 'type2')]

    def test___setitem___invalid_key(self):
        """
        Test setting an invalid attribute using __setitem__.
        This should raise an KeyError as the ObservationSpec class
        does not have a mechanism to handle arbitrary attribute creation.
        """
        obs_spec = ObservationSpec()
        with pytest.raises(KeyError):
            obs_spec['non_existent_attribute'] = 'some_value'

    def test_build_1(self):
        """
        Test the build method of ObservationSpec with invalid keys in simulation_parameters, 
        zone_simulation_parameters, and prediction_variables, along with various edge cases.

        This test covers the following scenarios:
        1. Invalid keys in simulation_parameters
        2. Invalid keys in zone_simulation_parameters
        3. Use of one-day weather prediction with invalid keys in prediction_variables
        4. Invalid prediction hours
        5. Non-empty variables, internal_variables, and meters
        6. Use of actuator state
        7. Ensuring at least one observation element is defined
        """
        obs_spec = ObservationSpec(
            variables=[("var1", "type1")],
            internal_variables=["int_var1"],
            meters=["meter1"],
            simulation_parameters={"invalid_key": True},
            zone_simulation_parameters={"invalid_zone_key": True},
            use_one_day_weather_prediction=True,
            # prediction_hours=25,
            prediction_variables={"invalid_pred_key": True},
            use_actuator_state=True
        )

        with pytest.raises(ValueError) as exc_info:
            obs_spec.build()

        assert "The key 'invalid_key' is not admissible in the simulation_parameters" in str(exc_info.value)

    def test_build_2(self):
        """
        Tests the build method of ObservationSpec class with various input conditions.

        This test covers the following scenarios:
        1. Valid simulation parameters
        2. Invalid zone simulation parameters
        3. Weather prediction enabled with invalid prediction variables
        4. Invalid prediction hours
        5. All observation types defined (variables, internal variables, meters, actuator state)
        6. Verifies that the method returns the object's attributes as a dictionary
        """
        obs_spec = ObservationSpec(
            variables=[("Zone Mean Air Temperature", "Zone 1")],
            internal_variables=["Zone Floor Area"],
            meters=["Electricity:Facility"],
            simulation_parameters={"day_of_week": True},
            zone_simulation_parameters={"invalid_key": True},
            use_one_day_weather_prediction=True,
            # prediction_hours=25,
            prediction_variables={"invalid_weather_var": True},
            use_actuator_state=True
        )

        with pytest.raises(ValueError, match="The key 'invalid_key' is not admissible in the zone_simulation_parameters"):
            obs_spec.build()

    def test_build_3(self):
        """
        Test the build method of ObservationSpec with various edge cases and conditions.

        This test checks:
        1. Invalid keys in simulation_parameters
        2. Valid keys in zone_simulation_parameters
        3. One-day weather prediction with invalid prediction variable
        4. Out-of-range prediction hours
        5. Presence of variables, internal variables, meters, and actuator state
        6. Ensuring at least one observation is defined

        Expected behavior:
        - Raises ValueError for invalid keys
        - Logs warning for out-of-range prediction hours
        - Successfully builds and returns the observation specification
        """
        obs_spec = ObservationSpec(
            variables=[("var1", "type1")],
            internal_variables=["int_var1"],
            meters=["meter1"],
            simulation_parameters={"invalid_key": True},
            zone_simulation_parameters={"zone_time_step": True},
            use_one_day_weather_prediction=True,
            # prediction_hours=25,
            prediction_variables={"invalid_weather_var": True},
            use_actuator_state=True,
            other_obs={"custom_obs": 1.0}
        )

        with pytest.raises(ValueError, match="The key 'invalid_key' is not admissible in the simulation_parameters"):
            obs_spec.build()

        # Correct the invalid key
        obs_spec.simulation_parameters = {"actual_date_time": True}

        with pytest.raises(ValueError, match="The key 'invalid_weather_var' is not admissible in the prediction_variables"):
            obs_spec.build()

        # Correct the invalid prediction variable
        obs_spec.prediction_variables = {"albedo": True}

        # Test with valid configuration
        result = obs_spec.build()

        assert isinstance(result, dict)
        # assert result["prediction_hours"] == PREDICTION_HOURS
        assert len(result["variables"]) > 0
        assert len(result["internal_variables"]) > 0
        assert len(result["meters"]) > 0
        assert result["use_actuator_state"] is True

    def test_build_4(self):
        """
        Test the build method of ObservationSpec when various conditions are met:
        - Invalid keys in simulation_parameters and zone_simulation_parameters
        - One-day weather prediction is used
        - Prediction hours are out of range
        - Various observation components are defined
        - No valid observation components are defined (counter == 0)

        Expected behavior:
        - Raises ValueError for invalid keys
        - Logs a warning for invalid prediction hours
        - Raises ValueError when no valid observation components are defined
        """
        obs_spec = ObservationSpec(
            variables=[("var1", "unit1")],
            internal_variables=["int_var1"],
            meters=["meter1"],
            simulation_parameters={"invalid_key": True},
            zone_simulation_parameters={"invalid_zone_key": True},
            use_one_day_weather_prediction=True,
            # prediction_hours=25,
            prediction_variables={"outdoor_dry_bulb": True},
            use_actuator_state=True,
            other_obs={"custom_obs": 1.0}
        )

        with pytest.raises(ValueError, match="The key 'invalid_key' is not admissible in the simulation_parameters"):
            obs_spec.build()

        # Reset simulation_parameters to a valid state
        obs_spec.simulation_parameters = SIMULATION_PARAMETERS.copy()

        with pytest.raises(ValueError, match="The key 'invalid_zone_key' is not admissible in the zone_simulation_parameters"):
            obs_spec.build()

        # Reset zone_simulation_parameters to a valid state
        obs_spec.zone_simulation_parameters = ZONE_SIMULATION_PARAMETERS.copy()

        # Test with all observation components set to None or empty
        obs_spec.variables = None
        obs_spec.internal_variables = None
        obs_spec.meters = None
        obs_spec.simulation_parameters = SIMULATION_PARAMETERS.copy()
        obs_spec.zone_simulation_parameters = ZONE_SIMULATION_PARAMETERS.copy()
        obs_spec.prediction_variables = PREDICTION_VARIABLES.copy()
        obs_spec.use_actuator_state = False
        obs_spec.other_obs = {}

        with pytest.raises(ValueError, match="At least one variable/meter/actuator/parameter must be defined in the observation"):
            obs_spec.build()

    def test_build_5(self):
        """
        Test the build method of ObservationSpec when multiple invalid conditions are met.

        This test checks the following conditions:
        1. Invalid keys in simulation_parameters
        2. Invalid keys in zone_simulation_parameters
        3. Use of one-day weather prediction with invalid keys
        4. Invalid prediction hours
        5. No variables, meters, or other valid observations defined

        Expected behavior: The method should raise a ValueError due to no valid observations being defined.
        """
        obs_spec = ObservationSpec(
            simulation_parameters={"invalid_key": True},
            zone_simulation_parameters={"invalid_zone_key": True},
            use_one_day_weather_prediction=True,
            prediction_variables={"invalid_prediction_key": True},
            # prediction_hours=25,
            internal_variables=[],
            meters=[],
            use_actuator_state=False
        )

        with pytest.raises(ValueError):
            obs_spec.build()

    def test_build_6(self):
        """
        Test the build method when invalid keys are provided for simulation_parameters,
        zone_simulation_parameters, and prediction_variables, and when prediction_hours
        is out of range. Also tests the case where no valid observation elements are defined.
        """
        obs_spec = ObservationSpec(
            variables=[('var1', 'type1')],
            meters=['meter1'],
            simulation_parameters={'invalid_key': True},
            zone_simulation_parameters={'invalid_key': True},
            use_one_day_weather_prediction=True,
            # prediction_hours=25,
            prediction_variables={'invalid_key': True},
            use_actuator_state=True
        )

        with pytest.raises(ValueError) as exc_info:
            obs_spec.build()

        assert "The key 'invalid_key' is not admissible in the simulation_parameters" in str(exc_info.value)

    def test_build_7(self):
        """
        Test the build method of ObservationSpec with invalid parameters and empty observation.

        This test checks the following conditions:
        - Invalid keys in simulation_parameters
        - Invalid keys in zone_simulation_parameters
        - Invalid keys in prediction_variables with use_one_day_weather_prediction set to True
        - Invalid prediction_hours
        - Empty observation (no variables, meters, or parameters defined)

        Expected behavior: The method should raise a ValueError due to empty observation.
        """
        obs_spec = ObservationSpec(
            variables=[],
            internal_variables=[],
            simulation_parameters={"invalid_key": True},
            zone_simulation_parameters={"invalid_key": True},
            use_one_day_weather_prediction=True,
            # prediction_hours=25,
            prediction_variables={"invalid_key": True},
            use_actuator_state=True
        )

        with pytest.raises(ValueError):
            obs_spec.build()

    def test_build_invalid_keys_and_empty_observation(self):
        """
        Test the build method with invalid keys in simulation parameters, zone simulation parameters,
        and prediction variables, along with an empty observation configuration.

        This test checks if the method correctly raises ValueError for invalid keys and empty observation,
        while also verifying the behavior when prediction_hours is out of range.
        """
        obs_spec = ObservationSpec(
            variables=[],
            internal_variables=[],
            meters=[],
            simulation_parameters={"invalid_key": True},
            zone_simulation_parameters={"invalid_zone_key": True},
            use_one_day_weather_prediction=True,
            # prediction_hours=25,
            prediction_variables={"invalid_prediction_key": True},
            use_actuator_state=False,
            other_obs={}
        )

        with pytest.raises(ValueError) as excinfo:
            obs_spec.build()

        assert "The key 'invalid_key' is not admissible in the simulation_parameters" in str(excinfo.value)

    def test_build_invalid_parameters(self):
        """
        Test the build method with invalid parameters to ensure it raises appropriate exceptions.

        This test covers the following scenarios:
        - Invalid keys in simulation_parameters
        - Invalid keys in zone_simulation_parameters
        - Invalid keys in prediction_variables when use_one_day_weather_prediction is True
        - Invalid prediction_hours (outside the range 1-24)
        - No variables/meters/actuators/parameters defined (counter == 0)
        """
        obs_spec = ObservationSpec(
            simulation_parameters={"invalid_key": True},
            zone_simulation_parameters={"invalid_zone_key": True},
            use_one_day_weather_prediction=True,
            prediction_variables={"invalid_prediction_key": True},
            # prediction_hours=25
        )

        with pytest.raises(ValueError) as excinfo:
            obs_spec.build()

        assert "The key 'invalid_key' is not admissible in the simulation_parameters" in str(excinfo.value)

        # Reset observation spec for the next test
        obs_spec = ObservationSpec(
            use_one_day_weather_prediction=True,
            # prediction_hours=0
        )

        with pytest.raises(ValueError) as excinfo:
            obs_spec.build()

        assert "At least one variable/meter/actuator/parameter must be defined in the observation" in str(excinfo.value)

    def test_build_invalid_simulation_parameters(self):
        """
        Test the build method when an invalid key is present in simulation_parameters.
        This test verifies that a ValueError is raised when an invalid key is provided.
        """
        obs_spec = ObservationSpec(
            variables=[("Zone Mean Air Temperature", "")],
            simulation_parameters={"invalid_key": True},
            zone_simulation_parameters={"invalid_key": True},
            use_one_day_weather_prediction=True,
            prediction_variables={"invalid_key": True},
            # prediction_hours=12,
            internal_variables=["Zone People Occupant Count"],
            meters=["Electricity:Facility"],
            use_actuator_state=True
        )

        with pytest.raises(ValueError) as excinfo:
            obs_spec.build()

        assert "is not admissible in the simulation_parameters" in str(excinfo.value)

    def test_empty_observation(self):
        """
        Test that an empty observation specification raises a ValueError.
        """
        with pytest.raises(ValueError) as excinfo:
            ObservationSpec().build()
        assert "At least one variable/meter/actuator/parameter must be defined in the observation" in str(excinfo.value)

    # def test_invalid_prediction_hours(self):
    #     """
    #     Test that an invalid prediction_hours value is corrected and a warning is logged.
    #     """
    #     obs_spec = ObservationSpec(variables=[("Site Outdoor Air Drybulb Temperature", "Environment")], use_one_day_weather_prediction=True, prediction_hours=25)
    #     result = obs_spec.build()
    #     assert result['prediction_hours'] == 24

    # def test_invalid_prediction_hours_2(self):
    #     """
    #     Test that the build method sets prediction_hours to the default value when an invalid value is provided.
    #     """
    #     obs_spec = ObservationSpec(variables=[("Site Outdoor Air Drybulb Temperature", "Environment")], use_one_day_weather_prediction=True, prediction_hours=25)
    #     result = obs_spec.build()
    #     assert result["prediction_hours"] == 24

    def test_invalid_prediction_variable_key(self):
        """
        Test that the build method raises a ValueError when an invalid key is provided in prediction_variables.
        """
        obs_spec = ObservationSpec(variables=[("Site Outdoor Air Drybulb Temperature", "Environment")], use_one_day_weather_prediction=True, prediction_variables={"invalid_key": True})
        with pytest.raises(ValueError) as exc_info:
            obs_spec.build()
        assert "The key 'invalid_key' is not admissible in the prediction_variables" in str(exc_info.value)

    def test_invalid_prediction_variables(self):
        """
        Test that an invalid key in prediction_variables raises a ValueError when use_one_day_weather_prediction is True.
        """
        with pytest.raises(ValueError) as excinfo:
            ObservationSpec(use_one_day_weather_prediction=True, prediction_variables={'invalid_key': True}).build()
        assert "The key 'invalid_key' is not admissible in the prediction_variables" in str(excinfo.value)

    def test_invalid_simulation_parameter_key(self):
        """
        Test that the build method raises a ValueError when an invalid key is provided in simulation_parameters.
        """
        obs_spec = ObservationSpec(simulation_parameters={"invalid_key": True})
        with pytest.raises(ValueError) as exc_info:
            obs_spec.build()
        assert "The key 'invalid_key' is not admissible in the simulation_parameters" in str(exc_info.value)

    def test_invalid_simulation_parameters(self):
        """
        Test that an invalid key in simulation_parameters raises a ValueError.
        """
        with pytest.raises(ValueError) as excinfo:
            ObservationSpec(simulation_parameters={'invalid_key': True}).build()
        assert "The key 'invalid_key' is not admissible in the simulation_parameters" in str(excinfo.value)

    def test_invalid_zone_simulation_parameter_key(self):
        """
        Test that the build method raises a ValueError when an invalid key is provided in zone_simulation_parameters.
        """
        obs_spec = ObservationSpec(zone_simulation_parameters={"invalid_key": True})
        with pytest.raises(ValueError) as exc_info:
            obs_spec.build()
        assert "The key 'invalid_key' is not admissible in the zone_simulation_parameters" in str(exc_info.value)

    def test_invalid_zone_simulation_parameters(self):
        """
        Test that an invalid key in zone_simulation_parameters raises a ValueError.
        """
        with pytest.raises(ValueError) as excinfo:
            ObservationSpec(zone_simulation_parameters={'invalid_key': True}).build()
        assert "The key 'invalid_key' is not admissible in the zone_simulation_parameters" in str(excinfo.value)

    def test_no_observation_defined(self):
        """
        Test that the build method raises a ValueError when no observation is defined.
        """
        obs_spec = ObservationSpec()
        with pytest.raises(ValueError) as exc_info:
            obs_spec.build()
        assert "At least one variable/meter/actuator/parameter must be defined in the observation" in str(exc_info.value)



"""
HVAC Systems actuators ActionMappers
======================================

System Node Setpoints
----------------------

A series of actuators is available for all the setpoints that can be placed on system nodes. System
nodes are used to define air and plant loops, and a natural application of EMS is to control the
setpoints at these nodes. The node actuators are all called “System Node Setpoint.” There are
nine control types:

* Temperature Setpoint, (°C)
* Temperature Minimum Setpoint (°C)
* Temperature Maximum Setpoint (°C)
* Humidity Ratio Setpoint (kgWater/kgDryAir)
* Humidity Ratio Minimum Setpoint (kgWater/kgDryAir)
* Humidity Ratio Maximum Setpoint (kgWater/kgDryAir)
* Mass Flow Rate Setpoint (kg/s)
* Mass Flow Rate Minimum Available Setpoint (kg/s)
* Mass Flow Rate Maximum Available Setpoint (kg/s)

Using these actuators is natural with an EMS. Typically, the controller would place the setpoint
on the outlet node. Then the component’s low-level controller should operate to meet the leaving
setpoint. Setting the setpoints on nodes should be a common application for the EMS.

Although all nine possible setpoints are available as EMS actuators, it does not follow that
EnergyPlus can use all of them. Most components can use only one or two setpoints. If a component
cannot control to meet the setpoints on a node, the actuator will do nothing.

"""
import gymnasium as gym
import numpy as np
from typing import Any, Dict, List, Tuple, Optional
from eprllib.Agents.ActionMappers.BaseActionMapper import BaseActionMapper
from eprllib.Utils.observation_utils import get_actuator_name
from eprllib.Utils.annotations import override

class SystemNodeSetpointsTemperatureSetpoint(BaseActionMapper):
    
    @override(BaseActionMapper)
    def setup(self):
        """
        Sets up the components of the module.

        This is called automatically during the __init__ method of this class.
        """
        # The name of the actuator that will be controlled by the agent. 
        # This will be obtained from the actuators_config in the environment configuration file.
        self.hvac_system_actuator: Optional[str] = None
        
        # Here we use the config dict to provide the range of the air mass flow rate.
        air_mass_flow_rate_range: Tuple[float|int, float|int] = self.action_mapper_config['variable_range']
        self.air_mass_flow_rate_range_max: float|int = max(air_mass_flow_rate_range)
        self.air_mass_flow_rate_range_min: float|int = min(air_mass_flow_rate_range)
        
    
    @override(BaseActionMapper)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Box(low=0.0, high=1.0, shape=(1,), dtype=np.float32)
    
    
    @override(BaseActionMapper)
    def actuator_names(
        self, 
        actuators_config: Dict[str, Tuple[str,str,str]]
        ) -> None:
        """
        This method is used to transform the agent dict action to actuator dict action. Consider that
        one agent could manage more than one actuator. For that reason, it is important to transform the
        action dict to actuator dict actions.

        Args:
            action (Any): The action to be transformed.
            actuators (List[str]): List of actuators controlled by the agent.

        Returns:
            Dict[str, Any]: Transformed actions for the actuators.
        """
        
        # Here the name of the actuator is obtained from the actuators_config 
        # in the environment configuration file.
        self.hvac_system_actuator = get_actuator_name(
            self.agent_name,
            actuators_config["HVAC System Actuator"][0],
            actuators_config["HVAC System Actuator"][1],
            actuators_config["HVAC System Actuator"][2]
        )
        
    
    @override(BaseActionMapper)
    def _agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str,Any]:
        """
        This method is used to transform the agent action to actuator dict action. Consider that
        one agent may manage more than one actuator.

        Args:
            action (Any): The agent action, normally an int of float.
            actuators (List[str]): The list of actuator names that the agent manage.

        Return:
            Dict[str, Any]: Dictionary with the actions for the actuators.
        """
        if self.hvac_system_actuator not in actuators:
            return {}
        else:
            if action <= 0.0:
                return {self.hvac_system_actuator: float(self.air_mass_flow_rate_range_min)}
            
            elif action >= 1.0:
                return {self.hvac_system_actuator: float(self.air_mass_flow_rate_range_max)}
            
            else:
                return {self.hvac_system_actuator: float(self.air_mass_flow_rate_range_min + \
                    (action) * (self.air_mass_flow_rate_range_max - self.air_mass_flow_rate_range_min))}



"""
Base Filter
============

This module contains the base class for defining ``Filter`` functions used in agent specifications.
``Filter``s are used to preprocess observations before they are fed to the agent. The ``BaseFilter``
class provides the basic structure and methods that can be extended to create custom ``Filter``s.

This class can not be used directly in ``eprllib``, but as a base to create new ```Filter``s. All the ``Filter``s
must be based in this class.

The methods provided here are used during inizialization and execution of the environment.
You have to overwrite the following methods:

    - ``setup(self)``
    - ``_get_filtered_obs``
    
"""
from typing import Any, Dict
from numpy import floating
from numpy.typing import NDArray

from eprllib import logger
from eprllib.Utils.annotations import OverrideToImplementCustomLogic

class BaseFilter:
    """
    Base class for defining filter functions used in agent specifications.
    Filters are used to preprocess observations before they are fed to the agent.
    """
    filter_fn_config: Dict[str, Any]
    agent_name: str
    _is_setup: bool
    
    def __init__(
        self,
        agent_name: str,
        filter_fn_config: Dict[str, Any] = {}
    ):
        """
        Initializes the BaseFilter class.

        Args:
        
            filter_fn_config (Dict[str, Any]): Configuration dictionary for the filter function.
            
        """
        self.filter_fn_config = filter_fn_config
        self.agent_name = agent_name
        
        logger.info(f"BaseFilter: The BaseFilter was correctly inicializated with {self.filter_fn_config} config.")
        
        # Make sure, `setup()` is only called once, no matter what.
        if hasattr(self, "_is_setup") and self._is_setup:
            raise RuntimeError(
                "``BaseActionMapper.setup()`` called twice within your ActionMapper implementation "
                f"{self}! Make sure you are using the proper inheritance order "
                " and that you are NOT overriding the constructor, but "
                "only the ``setup()`` method of your subclass."
            )
        try:
            self.setup()
        except AttributeError as e:
            raise e
        self._is_setup:bool = True
        
        # Check if the agent_name is defined.
        assert self.agent_name is not None, "BaseFilter: The agent name must be defined."
    
    
    def get_filtered_obs(
        self,
        agent_states: Dict[str, Any],
    ) -> NDArray[floating[Any]]:
        # Check if the agent_states dictionary is empty
        if not agent_states:
            msg = "agent_states dictionary is empty"
            logger.error(msg)
            raise ValueError(msg)
        
        # Check if all values in the agent_states dictionary are numeric
        if not all(isinstance(value, (int, float)) for value in agent_states.values()):
            msg = "All values in agent_states must be numeric"
            logger.error(msg)
            raise ValueError(msg)
        
        # Generate a copy of the agent_states to avoid conflicts with global variables.
        agent_states_copy = agent_states.copy()
        
        return self._get_filtered_obs(agent_states_copy)
    
    
    # ===========================
    # === OVERRIDABLE METHODS ===
    # ===========================
    
    @OverrideToImplementCustomLogic
    def setup(self):
        """
        Sets up the components of the module.

        This is called automatically during the __init__ method of this class.
        """
        pass
    
    
    @OverrideToImplementCustomLogic
    def _get_filtered_obs(
        self,
        agent_states: Dict[str, Any],
    ) -> NDArray[floating[Any]]:
        """
        Returns the filtered observations for the agent based on the environment configuration
        and agent states. This method processes the raw observations according to the filter
        configuration specified in filter_fn_config.

        Args:
        
            env_config (Dict[str, Any]): Configuration dictionary for the environment. Can include settings 
                that affect how the observations are filtered.
            agent_states (Dict[str, Any], optional): Dictionary containing the states of the agent.

        Returns:
        
            NDarray: Filtered observations as a numpy array of float64 values.
            
        Raises:
        
            NotImplementedError: If this method is not implemented in a subclass.
            
        """
        msg = "BaseFilter: This method should be implemented in a subclass."
        logger.error(msg)
        raise NotImplementedError(msg)


"""
This type stub file was generated by pyright.
"""

"""Benchmarks for utilities working with arbitrarily nested structures."""
TIME_UNITS = ...
def format_time(time): # -> str | None:
  ...

def run_benchmark(benchmark_fn, num_iters):
  ...

def map_to_list(func, *args): # -> list[Any]:
  ...

def benchmark_map(map_fn, structure): # -> Callable[[], Any]:
  ...

BENCHMARKS = ...
def main(): # -> None:
  ...

if __name__ == "__main__":
  ...



"""
This type stub file was generated by pyright.
"""

import logging
import sys
import wrapt
from collections import abc as collections_abc
from typing import Mapping, Sequence, TypeVar, Union
from .sequence import _is_attrs, _is_namedtuple, _sequence_like, _sorted
from tree import _tree

"""Functions for working with nested data structures."""
ObjectProxy = ...
__all__ = ["is_nested", "assert_same_structure", "unflatten_as", "flatten", "flatten_up_to", "flatten_with_path", "flatten_with_path_up_to", "map_structure", "map_structure_up_to", "map_structure_with_path", "map_structure_with_path_up_to", "traverse", "MAP_TO_NONE"]
__version__ = ...
_TEXT_OR_BYTES = ...
_SHALLOW_TREE_HAS_INVALID_KEYS = ...
_STRUCTURES_HAVE_MISMATCHING_TYPES = ...
_STRUCTURES_HAVE_MISMATCHING_LENGTHS = ...
_INPUT_TREE_SMALLER_THAN_SHALLOW_TREE = ...
_IF_SHALLOW_IS_SEQ_INPUT_MUST_BE_SEQ = ...
_IF_SHALLOW_IS_SEQ_INPUT_MUST_BE_SEQ_WITH_PATH = ...
K = TypeVar("K")
V = TypeVar("V")
StructureKV = Union[Sequence["StructureKV[K, V]"], Mapping[K, "StructureKV[K, V]"], V,]
Structure = StructureKV[str, V]
def is_nested(structure):
  """Checks if a given structure is nested.

  >>> tree.is_nested(42)
  False
  >>> tree.is_nested({"foo": 42})
  True

  Args:
    structure: A structure to check.

  Returns:
    `True` if a given structure is nested, i.e. is a sequence, a mapping,
    or a namedtuple, and `False` otherwise.
  """
  ...

def flatten(structure):
  r"""Flattens a possibly nested structure into a list.

  >>> tree.flatten([[1, 2, 3], [4, [5], [[6]]]])
  [1, 2, 3, 4, 5, 6]

  If `structure` is not nested, the result is a single-element list.

  >>> tree.flatten(None)
  [None]
  >>> tree.flatten(1)
  [1]

  In the case of dict instances, the sequence consists of the values,
  sorted by key to ensure deterministic behavior. This is true also for
  :class:`~collections.OrderedDict` instances: their sequence order is
  ignored, the sorting order of keys is used instead. The same convention
  is followed in :func:`~tree.unflatten`. This correctly unflattens dicts
  and ``OrderedDict``\ s after they have been flattened, and also allows
  flattening an ``OrderedDict`` and then unflattening it back using a
  corresponding plain dict, or vice-versa.

  Dictionaries with non-sortable keys cannot be flattened.

  >>> tree.flatten({100: 'world!', 6: 'Hello'})
  ['Hello', 'world!']

  Args:
    structure: An arbitrarily nested structure.

  Returns:
    A list, the flattened version of the input `structure`.

  Raises:
    TypeError: If `structure` is or contains a mapping with non-sortable keys.
  """
  ...

class _DotString:
  def __str__(self) -> str:
    ...
  
  def __repr__(self): # -> Literal['.']:
    ...
  


_DOT = ...
def assert_same_structure(a, b, check_types=...): # -> None:
  """Asserts that two structures are nested in the same way.

  >>> tree.assert_same_structure([(0, 1)], [(2, 3)])

  Note that namedtuples with identical name and fields are always considered
  to have the same shallow structure (even with `check_types=True`).

  >>> Foo = collections.namedtuple('Foo', ['a', 'b'])
  >>> AlsoFoo = collections.namedtuple('Foo', ['a', 'b'])
  >>> tree.assert_same_structure(Foo(0, 1), AlsoFoo(2, 3))

  Named tuples with different names are considered to have different shallow
  structures:

  >>> Bar = collections.namedtuple('Bar', ['a', 'b'])
  >>> tree.assert_same_structure(Foo(0, 1), Bar(2, 3))
  Traceback (most recent call last):
    ...
  TypeError: The two structures don't have the same nested structure.
  ...

  Args:
    a: an arbitrarily nested structure.
    b: an arbitrarily nested structure.
    check_types: if `True` (default) types of sequences are checked as
      well, including the keys of dictionaries. If set to `False`, for example
      a list and a tuple of objects will look the same if they have the same
      size. Note that namedtuples with identical name and fields are always
      considered to have the same shallow structure.

  Raises:
    ValueError: If the two structures do not have the same number of elements or
      if the two structures are not nested in the same way.
    TypeError: If the two structures differ in the type of sequence in any of
      their substructures. Only possible if `check_types` is `True`.
  """
  ...

def unflatten_as(structure, flat_sequence): # -> defaultdict[Any, Any] | MappingProxyType[Any, Any] | Mapping[Any, Any] | dict[Any, Any] | list[Any] | Any | ObjectProxy:
  r"""Unflattens a sequence into a given structure.

  >>> tree.unflatten_as([[1, 2], [[3], [4]]], [5, 6, 7, 8])
  [[5, 6], [[7], [8]]]

  If `structure` is a scalar, `flat_sequence` must be a single-element list;
  in this case the return value is ``flat_sequence[0]``.

  >>> tree.unflatten_as(None, [1])
  1

  If `structure` is or contains a dict instance, the keys will be sorted to
  pack the flat sequence in deterministic order. This is true also for
  :class:`~collections.OrderedDict` instances: their sequence order is
  ignored, the sorting order of keys is used instead. The same convention
  is followed in :func:`~tree.flatten`. This correctly unflattens dicts
  and ``OrderedDict``\ s after they have been flattened, and also allows
  flattening an ``OrderedDict`` and then unflattening it back using a
  corresponding plain dict, or vice-versa.

  Dictionaries with non-sortable keys cannot be unflattened.

  >>> tree.unflatten_as({1: None, 2: None}, ['Hello', 'world!'])
  {1: 'Hello', 2: 'world!'}

  Args:
    structure: Arbitrarily nested structure.
    flat_sequence: Sequence to unflatten.

  Returns:
    `flat_sequence` unflattened into `structure`.

  Raises:
    ValueError: If `flat_sequence` and `structure` have different
      element counts.
    TypeError: If `structure` is or contains a mapping with non-sortable keys.
  """
  ...

def map_structure(func, *structures, **kwargs): # -> object | defaultdict[Any, Any] | MappingProxyType[Any, Any] | Mapping[Any, Any] | dict[Any, Any] | list[Any] | Any | ObjectProxy:
  """Maps `func` through given structures.

  >>> structure = [[1], [2], [3]]
  >>> tree.map_structure(lambda v: v**2, structure)
  [[1], [4], [9]]
  >>> tree.map_structure(lambda x, y: x * y, structure, structure)
  [[1], [4], [9]]
  >>> Foo = collections.namedtuple('Foo', ['a', 'b'])
  >>> structure = Foo(a=1, b=2)
  >>> tree.map_structure(lambda v: v * 2, structure)
  Foo(a=2, b=4)

  Args:
    func: A callable that accepts as many arguments as there are structures.
    *structures: Arbitrarily nested structures of the same layout.
    **kwargs: The only valid keyword argument is `check_types`. If `True`
      (default) the types of components within the structures have
      to be match, e.g. ``tree.map_structure(func, [1], (1,))`` will raise
      a `TypeError`, otherwise this is not enforced. Note that namedtuples
      with identical name and fields are considered to be the same type.

  Returns:
    A new structure with the same layout as the given ones. If the
    `structures` have components of varying types, the resulting structure
    will use the same types as ``structures[0]``.

  Raises:
    TypeError: If `func` is not callable.
    ValueError: If the two structures do not have the same number of elements or
      if the two structures are not nested in the same way.
    TypeError: If `check_types` is `True` and any two `structures`
      differ in the types of their components.
    ValueError: If no structures were given or if a keyword argument other
      than `check_types` is provided.
  """
  ...

def map_structure_with_path(func, *structures, **kwargs): # -> defaultdict[Any, Any] | MappingProxyType[Any, Any] | Mapping[Any, Any] | dict[Any, Any] | list[Any] | Any | ObjectProxy:
  """Maps `func` through given structures.

  This is a variant of :func:`~tree.map_structure` which accumulates
  a *path* while mapping through the structures. A path is a tuple of
  indices and/or keys which uniquely identifies the positions of the
  arguments passed to `func`.

  >>> tree.map_structure_with_path(
  ...     lambda path, v: (path, v**2),
  ...     [{"foo": 42}])
  [{'foo': ((0, 'foo'), 1764)}]

  Args:
    func: A callable that accepts a path and as many arguments as there are
      structures.
    *structures: Arbitrarily nested structures of the same layout.
    **kwargs: The only valid keyword argument is `check_types`. If `True`
      (default) the types of components within the structures have to be match,
      e.g. ``tree.map_structure_with_path(func, [1], (1,))`` will raise a
      `TypeError`, otherwise this is not enforced. Note that namedtuples with
      identical name and fields are considered to be the same type.

  Returns:
    A new structure with the same layout as the given ones. If the
    `structures` have components of varying types, the resulting structure
    will use the same types as ``structures[0]``.

  Raises:
    TypeError: If `func` is not callable or if the `structures` do not
      have the same layout.
    TypeError: If `check_types` is `True` and any two `structures`
      differ in the types of their components.
    ValueError: If no structures were given or if a keyword argument other
      than `check_types` is provided.
  """
  ...

def flatten_up_to(shallow_structure, input_structure, check_types=...): # -> list[Any]:
  """Flattens `input_structure` up to `shallow_structure`.

  All further nested components in `input_structure` are retained as-is.

  >>> structure = [[1, 1], [2, 2]]
  >>> tree.flatten_up_to([None, None], structure)
  [[1, 1], [2, 2]]
  >>> tree.flatten_up_to([None, [None, None]], structure)
  [[1, 1], 2, 2]

  If `shallow_structure` and `input_structure` are not nested, the
  result is a single-element list:

  >>> tree.flatten_up_to(42, 1)
  [1]
  >>> tree.flatten_up_to(42, [1, 2, 3])
  [[1, 2, 3]]

  Args:
    shallow_structure: A structure with the same (but possibly more shallow)
      layout as `input_structure`.
    input_structure: An arbitrarily nested structure.
    check_types: If `True`, check that each node in shallow_tree has the
      same type as the corresponding node in `input_structure`.

  Returns:
    A list, the partially flattened version of `input_structure` wrt
    `shallow_structure`.

  Raises:
    TypeError: If the layout of `shallow_structure` does not match that of
      `input_structure`.
    TypeError: If `check_types` is `True` and `shallow_structure` and
      `input_structure` differ in the types of their components.
  """
  ...

def flatten_with_path_up_to(shallow_structure, input_structure, check_types=...): # -> list[tuple[tuple[()], Any] | tuple[tuple[()] | Any, Any]]:
  """Flattens `input_structure` up to `shallow_structure`.

  This is a combination of :func:`~tree.flatten_up_to` and
  :func:`~tree.flatten_with_path`

  Args:
    shallow_structure: A structure with the same (but possibly more shallow)
      layout as `input_structure`.
    input_structure: An arbitrarily nested structure.
    check_types: If `True`, check that each node in shallow_tree has the
      same type as the corresponding node in `input_structure`.

  Returns:
    A list of ``(path, item)`` pairs corresponding to the partially flattened
    version of `input_structure` wrt `shallow_structure`.

  Raises:
    TypeError: If the layout of `shallow_structure` does not match that of
      `input_structure`.
    TypeError: If `input_structure` is or contains a mapping with non-sortable
      keys.
    TypeError: If `check_types` is `True` and `shallow_structure` and
      `input_structure` differ in the types of their components.
  """
  ...

def map_structure_up_to(shallow_structure, func, *structures, **kwargs): # -> defaultdict[Any, Any] | MappingProxyType[Any, Any] | Mapping[Any, Any] | dict[Any, Any] | list[Any] | Any | ObjectProxy:
  """Maps `func` through given structures up to `shallow_structure`.

  This is a variant of :func:`~tree.map_structure` which only maps
  the given structures up to `shallow_structure`. All further nested
  components are retained as-is.

  >>> structure = [[1, 1], [2, 2]]
  >>> tree.map_structure_up_to([None, None], len, structure)
  [2, 2]
  >>> tree.map_structure_up_to([None, [None, None]], str, structure)
  ['[1, 1]', ['2', '2']]

  Args:
    shallow_structure: A structure with layout common to all `structures`.
    func: A callable that accepts as many arguments as there are structures.
    *structures: Arbitrarily nested structures of the same layout.
    **kwargs: No valid keyword arguments.
  Raises:
    ValueError: If `func` is not callable or if `structures` have different
      layout or if the layout of `shallow_structure` does not match that of
      `structures` or if no structures were given.

  Returns:
    A new structure with the same layout as `shallow_structure`.
  """
  ...

def map_structure_with_path_up_to(shallow_structure, func, *structures, **kwargs): # -> defaultdict[Any, Any] | MappingProxyType[Any, Any] | Mapping[Any, Any] | dict[Any, Any] | list[Any] | Any | ObjectProxy:
  """Maps `func` through given structures up to `shallow_structure`.

  This is a combination of :func:`~tree.map_structure_up_to` and
  :func:`~tree.map_structure_with_path`

  Args:
    shallow_structure: A structure with layout common to all `structures`.
    func: A callable that accepts a path and as many arguments as there are
      structures.
    *structures: Arbitrarily nested structures of the same layout.
    **kwargs: No valid keyword arguments.

  Raises:
    ValueError: If `func` is not callable or if `structures` have different
      layout or if the layout of `shallow_structure` does not match that of
      `structures` or if no structures were given.

  Returns:
    Result of repeatedly applying `func`. Has the same structure layout
    as `shallow_tree`.
  """
  ...

def flatten_with_path(structure): # -> list[tuple[tuple[()], Any] | tuple[tuple[()] | Any, Any]]:
  r"""Flattens a possibly nested structure into a list.

  This is a variant of :func:`~tree.flattens` which produces a list of
  pairs: ``(path, item)``.  A path is a tuple of indices and/or keys
  which uniquely identifies the position of the corresponding ``item``.

  >>> tree.flatten_with_path([{"foo": 42}])
  [((0, 'foo'), 42)]

  Args:
    structure: An arbitrarily nested structure.

  Returns:
    A list of ``(path, item)`` pairs corresponding to the flattened version
    of the input `structure`.

  Raises:
    TypeError:
      If ``structure`` is or contains a mapping with non-sortable keys.
  """
  ...

MAP_TO_NONE = ...
def traverse(fn, structure, top_down=...): # -> defaultdict[Any, Any] | MappingProxyType[Any, Any] | Mapping[Any, Any] | dict[Any, Any] | list[Any] | Any | ObjectProxy | None:
  """Traverses the given nested structure, applying the given function.

  The traversal is depth-first. If ``top_down`` is True (default), parents
  are returned before their children (giving the option to avoid traversing
  into a sub-tree).

  >>> visited = []
  >>> tree.traverse(visited.append, [(1, 2), [3], {"a": 4}], top_down=True)
  [(1, 2), [3], {'a': 4}]
  >>> visited
  [[(1, 2), [3], {'a': 4}], (1, 2), 1, 2, [3], 3, {'a': 4}, 4]

  >>> visited = []
  >>> tree.traverse(visited.append, [(1, 2), [3], {"a": 4}], top_down=False)
  [(1, 2), [3], {'a': 4}]
  >>> visited
  [1, 2, (1, 2), 3, [3], 4, {'a': 4}, [(1, 2), [3], {'a': 4}]]

  Args:
    fn: The function to be applied to each sub-nest of the structure.

      When traversing top-down:
        If ``fn(subtree) is None`` the traversal continues into the sub-tree.
        If ``fn(subtree) is not None`` the traversal does not continue into
        the sub-tree. The sub-tree will be replaced by ``fn(subtree)`` in the
        returned structure (to replace the sub-tree with None, use the special
        value :data:`MAP_TO_NONE`).

      When traversing bottom-up:
        If ``fn(subtree) is None`` the traversed sub-tree is returned unaltered.
        If ``fn(subtree) is not None`` the sub-tree will be replaced by
        ``fn(subtree)`` in the returned structure (to replace the sub-tree
        with None, use the special value :data:`MAP_TO_NONE`).

    structure: The structure to traverse.
    top_down: If True, parent structures will be visited before their children.

  Returns:
    The structured output from the traversal.
  """
  ...

def traverse_with_path(fn, structure, top_down=...): # -> defaultdict[Any, Any] | MappingProxyType[Any, Any] | Mapping[Any, Any] | dict[Any, Any] | list[Any] | Any | ObjectProxy | None:
  """Traverses the given nested structure, applying the given function.

  The traversal is depth-first. If ``top_down`` is True (default), parents
  are returned before their children (giving the option to avoid traversing
  into a sub-tree).

  >>> visited = []
  >>> tree.traverse_with_path(
  ...  lambda path, subtree: visited.append((path, subtree)),
  ...  [(1, 2), [3], {"a": 4}],
  ...  top_down=True)
  [(1, 2), [3], {'a': 4}]
  >>> visited == [
  ...  ((), [(1, 2), [3], {'a': 4}]),
  ...  ((0,), (1, 2)),
  ...  ((0, 0), 1),
  ...  ((0, 1), 2),
  ...  ((1,), [3]),
  ...  ((1, 0), 3),
  ...  ((2,), {'a': 4}),
  ...  ((2, 'a'), 4)]
  True

  >>> visited = []
  >>> tree.traverse_with_path(
  ...  lambda path, subtree: visited.append((path, subtree)),
  ...  [(1, 2), [3], {"a": 4}],
  ...  top_down=False)
  [(1, 2), [3], {'a': 4}]
  >>> visited == [
  ...  ((0, 0), 1),
  ...  ((0, 1), 2),
  ...  ((0,), (1, 2)),
  ...  ((1, 0), 3),
  ...  ((1,), [3]),
  ...  ((2, 'a'), 4),
  ...  ((2,), {'a': 4}),
  ... ((), [(1, 2), [3], {'a': 4}])]
  True

  Args:
    fn: The function to be applied to the path to each sub-nest of the structure
      and the sub-nest value.
      When traversing top-down: If ``fn(path, subtree) is None`` the traversal
        continues into the sub-tree. If ``fn(path, subtree) is not None`` the
        traversal does not continue into the sub-tree. The sub-tree will be
        replaced by ``fn(path, subtree)`` in the returned structure (to replace
        the sub-tree with None, use the special
        value :data:`MAP_TO_NONE`).
      When traversing bottom-up: If ``fn(path, subtree) is None`` the traversed
        sub-tree is returned unaltered. If ``fn(path, subtree) is not None`` the
        sub-tree will be replaced by ``fn(path, subtree)`` in the returned
        structure (to replace the sub-tree
        with None, use the special value :data:`MAP_TO_NONE`).
    structure: The structure to traverse.
    top_down: If True, parent structures will be visited before their children.

  Returns:
    The structured output from the traversal.
  """
  ...




# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import sys
import os

# Add the src directory to the path so we can import the version
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'src'))
from eprllib.version import __version__

project = 'eprllib'
copyright = '2026, Germán Rodolfo Henderson'
author = 'Germán Rodolfo Henderson'

release = __version__


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',  # To support Google and NumPy style docstrings
    'sphinx.ext.viewcode',  # Adds links to highlighted source code
    'sphinx.ext.githubpages',
]

templates_path = ['_templates']
exclude_patterns = ['Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_book_theme'
html_static_path = ['_static']
html_css_files = ["custom.css"]

html_theme_options = {
    "repository_url": "https://github.com/hermmanhender/eprllib",
    "repository_branch": "docs",
    "path_to_docs": "/docs",
    "use_edit_page_button": True,
    "use_repository_button": True,
    "use_download_button": True,
    "use_issues_button": True
}

html_logo = "Images/eprllib_logo.png"
html_title = "eprllib: use EnergyPlus as an environment for DRL control"
html_baseurl = "https://hermmanhender.github.io/eprllib/docs/"



"""
Annotations
============


"""
from eprllib import logger
from typing import Type, Callable, Any
from ray.tune.experiment.trial import Trial

def override(parent_cls: Type[Any]) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Decorator for documenting method overrides.

    Args:
        parent_cls: The superclass that provides the overridden method. If
            `parent_class` does not actually have the method or the class, in which
            method is defined is not a subclass of `parent_class`, an error is raised.
    Returns:
        A decorator that can be applied to methods in subclasses to indicate that
        they override a method from the parent class.
        
    Raises:
        TypeError: If the method does not exist in the parent class or if the class
            is not a subclass of the expected parent class.
        
    Example:
    .. testcode::
        :skipif: True

        from ray.rllib.policy import Policy
        class TorchPolicy(Policy):
            ...
            # Indicates that `TorchPolicy.loss()` overrides the parent
            # Policy class' own `loss method. Leads to an error if Policy
            # does not have a `loss` method.

            @override(Policy)
            def loss(self, model, action_dist, train_batch):
                ...

    """

    class OverrideCheck:
        def __init__(self, func: Callable[..., Any], expected_parent_cls: Type[Any]) -> None:
            self.func = func
            self.expected_parent_cls = expected_parent_cls

        def __set_name__(self, owner: Type[Any], name: str) -> None:
            # Check if the owner (the class) is a subclass of the expected base class
            if not issubclass(owner, self.expected_parent_cls):
                msg = (
                    f"Annotations: \n"
                    f"When using the @override decorator, {owner.__name__} must be a "
                    f"subclass of {parent_cls.__name__}!"
                )
                logger.error(msg)
                raise TypeError(msg)
            # Set the function as a regular method on the class.
            setattr(owner, name, self.func)

    def decorator(method: Callable[..., Any]) -> Callable[..., Any]:
        # Check, whether `method` is actually defined by the parent class.
        if method.__name__ not in dir(parent_cls):
            msg = (
                f"Annotations: \n"
                f"When using the @override decorator, {method.__name__} must override "
                f"the respective method (with the same name) of {parent_cls.__name__}!"
            )
            logger.error(msg)
            raise TypeError(msg)

        # Check if the class is a subclass of the expected base class
        OverrideCheck(method, parent_cls)
        return method

    return decorator


def OverrideToImplementCustomLogic(obj: Any) -> Any:
    """Users should override this in their sub-classes to implement custom logic.

    Used in ActionMappers, Filters, Rewards,and Connectors to tag methods that need overriding, e.g.
    ``ActionMappers.setup()``.
    """
    obj.__is_overridden__ = False
    return obj


def trial_str_creator_for_tune(trial: Trial, name:str='eprllib'):
    """
    This method create a description for the folder where the outputs and checkpoints 
    will be save.

    Args:
        trial: A trial type of RLlib.
        name (str): Optional name for the trial. Default: eprllib

    Returns:
        str: Return a unique string for the folder of the trial.
    """
    # Validar que trial tenga los atributos esperados
    if not (hasattr(trial, 'trainable_name') and hasattr(trial, 'trial_id')):
        msg = "TrialStrCreator: The 'trial' argument must have 'trainable_name' and 'trial_id' attributes."
        logger.error(msg)
        raise ValueError(msg)
    
    return "{}_{}_{}".format(name, trial.trainable_name, trial.trial_id)


"""
Specification for the ActionMapper functions
===========================================================
This module defines the ``ActionMapperSpec`` class, which is used to specify
the configuration of ActionMapper functions for agents in reinforcement 
learning environments.
"""
from typing import Dict, Any, Type, Optional
from eprllib.Agents.ActionMappers.BaseActionMapper import BaseActionMapper
from eprllib import logger

class ActionMapperSpec:
    """
    ActionMapperSpec is the base class for a ActionMapper specification to safe configuration of the object.
    """
    action_mapper_fn: Optional[Type[BaseActionMapper]] = None
    action_mapper_config: Dict[str, Any] = {}
    
    def __init__(
        self,
        action_mapper_fn: Optional[Type[BaseActionMapper]] = None,
        action_mapper_config: Dict[str, Any] = {},
    ) -> None:
        """
        Construction method.
        
        Args:
            action_mapper_fn (BaseActionMapper): The ActionMapper function takes the arguments agent_id, observation and returns the
            observation filtered. See ``eprllib.Agents.ActionMappers`` for examples.
            
            action_mapper_config (Dict[str, Any]): The configuration of the ActionMapper function.
        
        Returns:
            None
            
        Raises:
            ValueError: If the action_mapper is not defined.
            ValueError: If the action_mapper_config is not a dictionary.
            ValueError: If the action_mapper is not a subclass of BaseActionMapper.
        """
        self.action_mapper_fn = action_mapper_fn
        self.action_mapper_config = action_mapper_config
        
        logger.info(f"ActionMapperSpec: The ActionMapperSpec was correctly inicializated with {self.action_mapper_fn} class and {self.action_mapper_config} config.")
    
    def __getitem__(self, key:str) -> Any:
        return getattr(self, key)

    def __setitem__(self, key:str, value:Any) -> None:
        valid_keys = self.__dict__.keys()
        if key not in valid_keys:
            msg = f"ActionMapperSpec: Invalid key: {key}."
            logger.error(msg)
            raise KeyError(msg)
        setattr(self, key, value)
    
    def build(self) -> Dict[str, Any]:
        """
        This method is used to build the ActionMapperSpec object.
        """
        if self.action_mapper_fn == None:
            msg = "ActionMapperSpec: The ActionMapper class must be defined."
            logger.error(msg)
            raise ValueError(msg)
        
        logger.info(f"ActionMapperSpec: The ActionMapper class is {self.action_mapper_fn}.")
        
        return vars(self)



from eprllib.Agents.Filters.BaseFilter import BaseFilter
import pytest

class TestBasefilter:

    def test___init___1(self):
        """
        Test the initialization of BaseFilter with a valid filter_fn_config.

        This test verifies that the BaseFilter constructor correctly initializes
        the filter_fn_config attribute with the provided configuration dictionary.
        """
        filter_fn_config = {"key": "value"}
        base_filter = BaseFilter("agent_name",filter_fn_config)
        assert base_filter.filter_fn_config == filter_fn_config

    # def test_get_filtered_obs_1(self):
    #     """
    #     Test that get_filtered_obs raises NotImplementedError in BaseFilter.

    #     This test verifies that calling get_filtered_obs on a BaseFilter instance
    #     raises a NotImplementedError with the expected message, as the method
    #     should be implemented in subclasses.
    #     """
    #     base_filter = BaseFilter({})
    #     env_config = {}
    #     agent_states = {}

    #     with pytest.raises(NotImplementedError) as excinfo:
    #         base_filter.get_filtered_obs(env_config, agent_states)

    #     assert str(excinfo.value) == "BaseFilter: This method should be implemented in a subclass."

    # def test_get_filtered_obs_not_implemented(self):
    #     """
    #     Test that calling get_filtered_obs on BaseFilter raises NotImplementedError.
    #     This is the only edge case explicitly handled in the focal method's implementation.
    #     """
    #     base_filter = BaseFilter({})
    #     with pytest.raises(NotImplementedError) as excinfo:
    #         base_filter.get_filtered_obs({}, {})
    #     assert str(excinfo.value) == "BaseFilter: This method should be implemented in a subclass."


    def test_init_with_empty_dict(self):
        """
        Test initialization of BaseFilter with an empty dictionary.
        This tests the edge case where the filter_fn_config is an empty dictionary,
        which is a valid input but may lead to unexpected behavior in subclasses.
        """
        filter_instance = BaseFilter("agent_name",{})
        assert filter_instance.filter_fn_config == {}

    # def test_init_with_none_input(self):
    #     """
    #     Test initialization of BaseFilter with None as input.
    #     This tests the edge case where the filter_fn_config is None, which is not
    #     a valid dictionary input as required by the method signature.
    #     """
    #     with pytest.raises(TypeError) as excinfo:
    #         BaseFilter(None)
    #     assert str(excinfo.value) == "filter_fn_config must be a dictionary"


"""
Environments
============

The environments in RLlib can be implemented as single-agent, multi-agent, or
external environments. In ``eprllib``, we use a multi-agent approach, allowing multiple 
agents to run in the environment, but also supporting a single agent if only one is specified.

The standard configuration uses a policy with fully shared parameters, but future 
versions aim to add flexibility to the policy.

You can configure the environment with the :class:`~eprllib.Environment.EnvironmentConfig.EnvironmentConfig` class.

The module includes the following classes and functions:

- ``Environment``: The base class for creating multi-agent environments.
- ``EnvironmentRunner``: The base class for running EnergyPlus simulations.
- ``EnvironmentConfig``: The class used to configure the environment.
"""

TIMEOUT = 10.0
CUT_EPISODE_LEN = 0



"""
Filter Utilities
===================

Work in progress...
"""

import numpy as np

def to_sin_transformation(
    value: float,
    min: float,
    max: float
    ) -> float:
    """
    Transform a value to a sin function.
    
    Args:
        value (float): The value to transform.
        min (float): The minimum value of the range.
        max (float): The maximum value of the range.
        
    Returns:
        float: The transformed value.
    """
    return np.sin(2 * np.pi * (value - min) / (max - min))

def normalization_minmax(
    value: float, 
    min: float, 
    max: float,
    a: float = -1.0,
    b: float = 1.0
    ) -> float:
    """
    Normalize a value between a and b.
    
    Args:
        value (float): The value to normalize.
        min (float): The minimum value of the range.
        max (float): The maximum value of the range.
        a (float, optional): The lower bound of the normalization range. Defaults to -1.0.
        b (float, optional): The upper bound of the normalization range. Defaults to 1.0.
        
    Returns:
        float: The normalized value.
    """
    return a + (b - a) * (value - min) / (max - min)

def desnormalization_minmax(
    value: float, 
    min: float, 
    max: float,
    a: float = -1.0,
    b: float = 1.0
    ) -> float:
    """
    Desnormalize a value between a and b.
    
    Args:
        value (float): The value to normalize.
        min (float): The minimum value of the range.
        max (float): The maximum value of the range.
        a (float, optional): The lower bound of the normalization range. Defaults to -1.0.
        b (float, optional): The upper bound of the normalization range. Defaults to 1.0.
        
    Returns:
        float: The desnormalized value.
    """
    return min + (max - min) * (value - a) / (b - a)


from eprllib.Agents.Filters.BaseFilter import BaseFilter
from eprllib.Agents.Filters.FilterSpec import FilterSpec

import logging
import pytest

logger = logging.getLogger("ray.rllib")

class TestFilterspec:

    def test___getitem___1(self):
        """
        Test that the __getitem__ method correctly returns the value of the specified attribute.
        """
        filter_spec = FilterSpec(filter_fn=BaseFilter, filter_fn_config={'test_config': 'value'})
        assert filter_spec['filter_fn'] == BaseFilter
        assert filter_spec['filter_fn_config'] == {'test_config': 'value'}

    def test___getitem___nonexistent_attribute(self):
        """
        Test that accessing a non-existent attribute raises an AttributeError.
        """
        filter_spec = FilterSpec()
        with pytest.raises(AttributeError):
            filter_spec['nonexistent_key']

    def test___init___default_parameters(self):
        """
        Test the __init__ method of FilterSpec with default parameters.

        This test verifies that when FilterSpec is initialized without any arguments,
        it correctly sets the default values for filter_fn and filter_fn_config.
        """
        filter_spec = FilterSpec()
        assert filter_spec.filter_fn is None
        assert filter_spec.filter_fn_config == {}

    def test___setitem___2(self):
        """
        Test that __setitem__ successfully sets a valid key-value pair.

        This test verifies that the __setitem__ method of FilterSpec correctly
        sets a value for an existing key without raising any exceptions.
        """
        filter_spec = FilterSpec()
        filter_spec['filter_fn'] = BaseFilter

        assert filter_spec['filter_fn'] == BaseFilter

    def test___setitem___invalid_key(self):
        """
        Test the __setitem__ method with an invalid key.
        This should raise a KeyError as explicitly handled in the method.
        """
        filter_spec = FilterSpec()
        with pytest.raises(KeyError) as excinfo:
            filter_spec['invalid_key'] = 'some_value'
        assert str(excinfo.value) == "'FilterSpec: Invalid key: invalid_key.'"

    def test___setitem___invalid_key_2(self):
        """
        Test that __setitem__ raises a KeyError when an invalid key is provided.
        This test covers the path where the provided key is not in the valid keys.
        """
        filter_spec = FilterSpec()
        with pytest.raises(KeyError) as exc_info:
            filter_spec['invalid_key'] = 'some_value'
        assert str(exc_info.value) == "'FilterSpec: Invalid key: invalid_key.'"




"""
Filters
========

This module contains classes for defining and applying filters to preprocess observations before they are fed to the agent.
Filters are essential for modifying the raw observations from the environment to a format that is suitable for the agent's
policy.

In this module, you will find:

- ``BaseFilter``: The base class for defining filter functions. It provides the basic structure and methods that can be 
extended to create custom filters.
- ``FilterSpec``: A class used to specify the configuration of filter functions in the agent definition. It ensures that 
the filter function is properly defined and adheres to the expected interface.

These filters can be used to preprocess observations in various ways, depending on the specific requirements of the 
agent's policy.
"""


import shutil
import json
import platform
import multiprocessing
from pathlib import Path
from typing import Dict, Optional, Any
from eprllib import logger

def parallel_energyplus_setup(ep_base_path: str, num_workers: Optional[int] = None) -> Dict[int, str]:
    """
    Verifica y crea copias del directorio de EnergyPlus para ejecución en paralelo por cada Worker.
    
    :param ep_base_path: Ruta original de EnergyPlus (ej. '/usr/local/EnergyPlus-24-2-0').
    :param num_workers: Cantidad de Rollout Workers configurados en RLlib.
    :return: Diccionario mapeando {worker_id: ruta_al_directorio_copia}.
    """
    if num_workers is None:
        num_workers = multiprocessing.cpu_count()
        
    ep_path = Path(ep_base_path)
    if not ep_path.exists():
        msg = f"El directorio base de EnergyPlus no existe: {ep_path}"
        logger.error(msg)
        raise FileNotFoundError(msg)
        
    # Archivo de control para evitar recrear los directorios innecesariamente
    archivo_json = ep_path / "parallel_setup_workers.json"
    rutas_workers: Dict[int, str] = {}
    
    # 1. Comprobar si las copias ya están creadas
    if archivo_json.exists():
        try:
            with open(archivo_json, "r") as f:
                datos = json.load(f)
                if datos.get("num_workers", 0) >= num_workers:
                    print("Las copias de EnergyPlus ya existen. Cargando rutas...")
                    # Convertimos de nuevo las llaves de string a int (JSON guarda las llaves como string)
                    return {int(k): v for k, v in datos.get("rutas_workers", {}).items()}
        except Exception as e:
            msg = f"Error leyendo el JSON de configuración. Recreando directorios. Detalles: {e}"
            logger.warning(msg)

    logger.info(f"Generando directorios de EnergyPlus para {num_workers} workers...")
    
    # 2. Identificar archivos necesarios considerando el SO
    sistema = platform.system()
    
    # Archivos comunes a todos los sistemas
    archivos_base = ["Energy+.idd", "pyenergyplus"]
    
    if sistema == "Windows":
        archivos_os = [
            "EnergyPlus.exe", "EPMacro.exe", "ExpandObjects.exe",
            "DElight2.dll", "libexpat.dll", "bcvtb.dll"
        ]
    elif sistema == "Linux":
        archivos_os = [
            "energyplus", "EPMacro", "ExpandObjects",
            "libDElight2.so", "libbcvtb.so",
            "libexpat.so", "libexpat.so.1" # A veces Linux incluye la versión en el nombre
        ]
    elif sistema == "Darwin":  # MacOS
        archivos_os = [
            "energyplus", "EPMacro", "ExpandObjects",
            "libDElight2.dylib", "libbcvtb.dylib",
            "libexpat.dylib", "libexpat.1.dylib" 
        ]
    else:
        raise OSError(f"Sistema operativo no soportado: {sistema}")
    
    archivos_requeridos = archivos_base + archivos_os
    
    # 3. Crear directorios (sumamos 1 porque RLlib tiene un Local Worker con índice 0 y N Remote Workers)
    for worker_id in range(num_workers + 1):
        nombre_dir = f"{ep_path.name}_worker_{worker_id}"
        worker_dir = ep_path.parent / nombre_dir
        worker_dir.mkdir(parents=True, exist_ok=True)
        rutas_workers[worker_id] = str(worker_dir)
        
        for archivo in archivos_requeridos:
            origen = ep_path / archivo
            destino = worker_dir / archivo
            
            # Usamos exist() porque pusimos variaciones de libexpat. Copiará la que encuentre.
            if origen.exists():
                if origen.is_dir() and not destino.exists():
                    shutil.copytree(origen, destino)
                elif not origen.is_dir():
                    shutil.copy2(origen, destino)
            else:
                # Opcional: Avisar si falta un archivo crítico, aunque con las variaciones de libexpat es normal que falte alguna
                if archivo in ["energyplus", "EnergyPlus.exe", "Energy+.idd", "pyenergyplus"]:
                    logger.warning(f"ADVERTENCIA: No se encontró el archivo crítico '{archivo}' en {ep_path}")
                    
        # 4. Asegurar que Energy+.ini esté vacío
        with open(worker_dir / "Energy+.ini", "w") as f:
            f.write("")
            
    # 5. Registrar el estado en el JSON
    config: Dict[str, Any]= {
        "num_workers": num_workers,
        "base_path": str(ep_path),
        "rutas_workers": rutas_workers
    }
    with open(archivo_json, "w") as f:
        json.dump(config, f, indent=4)
        
    logger.info("Directorios generados exitosamente.")
    
    return rutas_workers



from eprllib.Utils.agent_utils import config_validation
from typing import Tuple
import pytest

class TestAgentUtils:

    def test_config_validation_2(self):
        """
        Tests config_validation function when a key is present, the expected type is a tuple,
        but the value is not a tuple, has incorrect number of elements, or contains an item
        of incorrect type.
        """
        config = {
            'test_key': [1, 'string']
        }
        required_keys = {
            'test_key': Tuple[int, int, str]
        }

        with pytest.raises(TypeError) as exc_info:
            config_validation(config, required_keys)

        assert str(exc_info.value) == "The key 'test_key' must be a tuple, but got a list"

        # Change config to have a tuple with incorrect number of elements
        config['test_key'] = (1, 'string')

        with pytest.raises(ValueError) as exc_info:
            config_validation(config, required_keys)

        assert str(exc_info.value) == "The key 'test_key' must be a tuple with 3 elements, but has 2"

        # Change config to have a tuple with correct number of elements but incorrect type
        config['test_key'] = (1, 2, 2)

        with pytest.raises(TypeError) as exc_info:
            config_validation(config, required_keys)

        assert str(exc_info.value) == "The element 2 in key 'test_key' must be a str, but got a int"

    def test_config_validation_incorrect_tuple_length(self):
        """
        Test config_validation when a tuple value in the config has an incorrect length.
        This should raise a ValueError.
        """
        config = {"key1": (1, 2, 3)}
        required_keys = {"key1": Tuple[int, int]}
        with pytest.raises(ValueError) as context:
            config_validation(config, required_keys)
            
        assert str(context.value) == "The key 'key1' must be a tuple with 2 elements, but has 3"

    def test_config_validation_incorrect_tuple_type(self):
        """
        Test config_validation when a tuple value in the config has an incorrect type.
        This should raise a TypeError.
        """
        config = {"key1": (1, "two")}
        required_keys = {"key1": Tuple[int, int]}
        with pytest.raises(TypeError) as context:
            config_validation(config, required_keys)
        
        assert str(context.value) == "The element 1 in key 'key1' must be a int, but got a str"

    def test_config_validation_incorrect_type(self):
        """
        Test config_validation when a value in the config has an incorrect type.
        This should raise a TypeError.
        """
        config = {"key1": "value1", "key2": "not_an_int"}
        required_keys = {"key1": str, "key2": int}
        with pytest.raises(TypeError) as context:
            config_validation(config, required_keys)
        
        assert str(context.value) == "The key 'key2' expects the type int, but got str"

    def test_config_validation_missing_key(self):
        """
        Test config_validation when a required key is missing from the config.
        This test covers the path where a key is not in the config,
        and the code raises a ValueError.
        """
        config = {"key1": "value1"}
        required_keys = {"key1": str, "key2": int}

        with pytest.raises(ValueError) as excinfo:
            config_validation(config, required_keys)

        assert "Missing required key: 'key2'" in str(excinfo.value)

    def test_config_validation_missing_key_2(self):
        """
        Test config_validation when a required key is missing from the config.
        This should raise a ValueError.
        """
        config = {"key1": "value1"}
        required_keys = {"key1": str, "key2": int}
        with pytest.raises(ValueError) as context:
            config_validation(config, required_keys)
        
        assert str(context.value) == "Missing required key: 'key2'"

    def test_config_validation_missing_key_and_tuple_mismatch(self):
        """
        Test config_validation when a key is missing, and a tuple value doesn't match the expected type.

        This test covers the following scenarios:
        1. A required key is missing from the config
        2. A value is a tuple when expected
        3. The tuple's length doesn't match the expected length
        4. The tuple's elements have the correct types
        """
        config = {
            'existing_key': 'value',
            'tuple_key': (1, 'string', True)
        }
        required_keys = {
            'missing_key': str,
            'tuple_key': Tuple[int, str, bool, float]
        }

        with pytest.raises(ValueError) as excinfo:
            config_validation(config, required_keys)
        assert "Missing required key: 'missing_key'" in str(excinfo.value)

        # Update config to include the missing key
        config['missing_key'] = 'value'

        with pytest.raises(ValueError) as excinfo:
            config_validation(config, required_keys)
        assert "The key 'tuple_key' must be a tuple with 4 elements, but has 3" in str(excinfo.value)

    def test_config_validation_missing_key_and_tuple_type_mismatch(self):
        """
        Test config_validation with missing key, tuple type mismatch, and incorrect tuple element type.

        This test covers the following scenarios:
        1. A required key is missing from the config.
        2. A value that should be a tuple is not a tuple.
        3. A tuple value has an incorrect number of elements.
        4. An element in a tuple has an incorrect type.
        """
        config = {
            'existing_key': 'value',
            'tuple_key': (1, 'string'),
            'wrong_tuple_key': [1, 2, 3],
            'wrong_tuple_length': (1, 2),
            'wrong_tuple_type': (1, 2, 'string')
        }
        required_keys = {
            'missing_key': str,
            'tuple_key': Tuple[int, str],
            'wrong_tuple_key': Tuple[int, int],
            'wrong_tuple_length': Tuple[int, int, int],
            'wrong_tuple_type': Tuple[int, int, int]
        }

        with pytest.raises(ValueError, match="Missing required key: 'missing_key'"):
            config_validation(config, required_keys)

        config['missing_key'] = 'value'
        with pytest.raises(TypeError, match="The key 'wrong_tuple_key' must be a tuple, but got a list"):
            config_validation(config, required_keys)

        config['wrong_tuple_key'] = (1, 2)
        with pytest.raises(ValueError, match="The key 'wrong_tuple_length' must be a tuple with 3 elements, but has 2"):
            config_validation(config, required_keys)

        config['wrong_tuple_length'] = (1, 2, 3)
        with pytest.raises(TypeError, match="The element 2 in key 'wrong_tuple_type' must be a int, but got a str"):
            config_validation(config, required_keys)

    def test_config_validation_missing_key_and_wrong_type(self):
        """
        Test config_validation when a required key is missing and another key has an incorrect type.

        This test covers the scenario where:
        1. A required key is not present in the config dictionary.
        2. Another key is present but its value has an incorrect type.

        The test expects ValueError to be raised for the missing key.
        """
        config = {
            'existing_key': 'string_value'
        }
        required_keys = {
            'missing_key': int,
            'existing_key': int
        }

        with pytest.raises(ValueError) as excinfo:
            config_validation(config, required_keys)

        assert "Missing required key: 'missing_key'" in str(excinfo.value)

    def test_config_validation_non_tuple_for_tuple_type(self):
        """
        Test config_validation when a non-tuple value is provided for a tuple type.
        This should raise a TypeError.
        """
        config = {"key1": "not_a_tuple"}
        required_keys = {"key1": Tuple[int, int]}
        with pytest.raises(TypeError) as context:
            config_validation(config, required_keys)
        
        assert str(context.value) == "The key 'key1' must be a tuple, but got a str"

    def test_config_validation_tuple_type_mismatch(self):
        """
        Test config_validation when a tuple value contains an element of incorrect type.

        This test covers the following path:
        - The key exists in the config
        - The expected type is a tuple
        - The value is a tuple
        - The tuple length matches the expected length
        - One element in the tuple has an incorrect type
        """
        config = {
            'test_key': (1, 'string', 3.14)
        }
        required_keys = {
            'test_key': Tuple[int, str, int]
        }

        with pytest.raises(TypeError) as context:
            config_validation(config, required_keys)

        assert str(context.value) == "The element 2 in key 'test_key' must be a int, but got a float"

    def test_config_validation_tuple_type_mismatch_2(self):
        """
        Test config_validation when a key is missing, the expected type is a tuple,
        the value is a tuple but with incorrect length, and contains an item of incorrect type.
        """
        config = {
            'key1': (1, 'string', 3.14)
        }
        required_keys = {
            'key1': Tuple[int, str, float],
            'missing_key': str
        }

        with pytest.raises(ValueError) as excinfo:
            config_validation(config, required_keys)
        assert "Missing required key: 'missing_key'" in str(excinfo.value)

        config['missing_key'] = 'value'
        config['key1'] = (1, 'string', '3.14', 4)  # Incorrect length and type

        with pytest.raises(ValueError) as excinfo:
            config_validation(config, required_keys)
        assert "The key 'key1' must be a tuple with 3 elements, but has 4" in str(excinfo.value)

        config['key1'] = (1, 'string', '3.14')  # Correct length but incorrect type for last element

        with pytest.raises(TypeError) as excinfo:
            config_validation(config, required_keys)
        assert "The element 2 in key 'key1' must be a float, but got a str" in str(excinfo.value)



# """
# Test EnvironmentRunner
# =====================

# This module contains tests for the EnvironmentRunner class.
# """
# import os
# import pytest
# import tempfile
# from unittest.mock import MagicMock, patch, call
# from queue import Queue
# from eprllib.Environment.EnvironmentRunner import EnvironmentRunner


# class TestEnvironmentRunner:
#     """Test class for EnvironmentRunner."""

#     @pytest.fixture
#     def mock_env_config(self):
#         """Create a mock environment configuration for testing."""
#         return {
#             "epjson_path": "tests/data/1ZoneDataCenterCRAC_wApproachTemp.idf",
#             "epw_path": "tests/data/USA_CA_San.Francisco.Intl.AP.724940_TMY3.epw",
#             "output_path": tempfile.gettempdir(),
#             "ep_terminal_output": False,
#             "timeout": 10,
#             "evaluation": False,
#             "agents_config": {
#                 "agent1": {
#                     "observation": {
#                         "variables": None,
#                         "internal_variables": None,
#                         "meters": None,
#                         "simulation_parameters": {
#                             "hour": True,
#                             "current_time": True,
#                         },
#                         "zone_simulation_parameters": {
#                             "system_time_step": False,
#                         },
#                         "use_one_day_weather_prediction": False,
#                         "prediction_hours": 0,
#                         "prediction_variables": {},
#                         "other_obs": {}
#                     },
#                     "action": {
#                         "actuators": [
#                             ["Schedule:Constant", "Schedule Value", "HTGSETP_SCH"]
#                         ]
#                     }
#                 }
#             }
#         }

#     @pytest.fixture
#     def mock_queues(self):
#         """Create mock queues for testing."""
#         return {
#             "obs_queue": MagicMock(spec=Queue),
#             "act_queue": MagicMock(spec=Queue),
#             "infos_queue": MagicMock(spec=Queue)
#         }

#     @pytest.fixture
#     def mock_functions(self):
#         """Create mock functions for testing."""
#         return {
#             "filter_fn": {"agent1": MagicMock()},
#             "action_mapper_fn": {"agent1": MagicMock()},
#             "connector_fn": MagicMock()
#         }

#     @patch('eprllib.Environment.EnvironmentRunner.api')
#     def test_initialization(self, mock_api, mock_env_config, mock_queues, mock_functions):
#         """Test initialization of EnvironmentRunner."""
#         # Create an instance of EnvironmentRunner
#         runner = EnvironmentRunner(
#             env_config=mock_env_config,
#             episode=0,
#             obs_queue=mock_queues["obs_queue"],
#             act_queue=mock_queues["act_queue"],
#             infos_queue=mock_queues["infos_queue"],
#             agents=["agent1"],
#             filter_fn=mock_functions["filter_fn"],
#             action_mapper_fn=mock_functions["action_mapper_fn"],
#             connector_fn=mock_functions["connector_fn"]
#         )
        
#         # Check that the runner was initialized correctly
#         assert runner.env_config == mock_env_config
#         assert runner.episode == 0
#         assert runner.obs_queue == mock_queues["obs_queue"]
#         assert runner.act_queue == mock_queues["act_queue"]
#         assert runner.infos_queue == mock_queues["infos_queue"]
#         assert runner.agents == ["agent1"]
#         assert runner.filter_fn == mock_functions["filter_fn"]
#         assert runner.action_mapper_fn == mock_functions["action_mapper_fn"]
#         assert runner.connector_fn == mock_functions["connector_fn"]
#         assert runner.energyplus_exec_thread is None
#         assert runner.energyplus_state is None
#         assert runner.sim_results == 0
#         assert runner.initialized is False
#         assert runner.init_handles is False
#         assert runner.simulation_complete is False
#         assert runner.first_observation is True
#         assert isinstance(runner.obs, dict)
#         assert isinstance(runner.infos, dict)
#         assert "agent1" in runner.infos
#         assert runner.is_last_timestep is False

#     @patch('eprllib.Environment.EnvironmentRunner.api')
#     @patch('eprllib.Environment.EnvironmentRunner.threading')
#     def test_start(self, mock_threading, mock_api, mock_env_config, mock_queues, mock_functions):
#         """Test the start method."""
#         # Create an instance of EnvironmentRunner
#         runner = EnvironmentRunner(
#             env_config=mock_env_config,
#             episode=0,
#             obs_queue=mock_queues["obs_queue"],
#             act_queue=mock_queues["act_queue"],
#             infos_queue=mock_queues["infos_queue"],
#             agents=["agent1"],
#             filter_fn=mock_functions["filter_fn"],
#             action_mapper_fn=mock_functions["action_mapper_fn"],
#             connector_fn=mock_functions["connector_fn"]
#         )
        
#         # Mock the thread
#         mock_thread = MagicMock()
#         mock_threading.Thread.return_value = mock_thread
        
#         # Call start
#         runner.start()
        
#         # Check that the EnergyPlus state was created
#         mock_api.state_manager.new_state.assert_called_once()
        
#         # Check that callbacks were registered
#         assert mock_api.runtime.callback_begin_zone_timestep_after_init_heat_balance.called
#         assert mock_api.runtime.callback_end_zone_timestep_after_zone_reporting.called
#         assert mock_api.runtime.callback_progress.called
        
#         # Check that the thread was started
#         mock_thread.start.assert_called_once()
#         assert runner.energyplus_exec_thread == mock_thread

#     @patch('eprllib.Environment.EnvironmentRunner.api')
#     def test_stop(self, mock_api, mock_env_config, mock_queues, mock_functions):
#         """Test the stop method."""
#         # Create an instance of EnvironmentRunner
#         runner = EnvironmentRunner(
#             env_config=mock_env_config,
#             episode=0,
#             obs_queue=mock_queues["obs_queue"],
#             act_queue=mock_queues["act_queue"],
#             infos_queue=mock_queues["infos_queue"],
#             agents=["agent1"],
#             filter_fn=mock_functions["filter_fn"],
#             action_mapper_fn=mock_functions["action_mapper_fn"],
#             connector_fn=mock_functions["connector_fn"]
#         )
        
#         # Mock the thread
#         runner.energyplus_exec_thread = MagicMock()
#         runner.energyplus_state = MagicMock()
        
#         # Call stop
#         runner.stop()
        
#         # Check that the simulation was stopped
#         mock_api.runtime.stop_simulation.assert_called_once_with(runner.energyplus_state)
        
#         # Check that the state was deleted
#         mock_api.state_manager.delete_state.assert_called_once_with(runner.energyplus_state)
        
#         # Check that the thread was reset
#         assert runner.energyplus_exec_thread is None
#         assert runner.first_observation is True

#     @patch('eprllib.Environment.EnvironmentRunner.api')
#     def test_failed(self, mock_api, mock_env_config, mock_queues, mock_functions):
#         """Test the failed method."""
#         # Create an instance of EnvironmentRunner
#         runner = EnvironmentRunner(
#             env_config=mock_env_config,
#             episode=0,
#             obs_queue=mock_queues["obs_queue"],
#             act_queue=mock_queues["act_queue"],
#             infos_queue=mock_queues["infos_queue"],
#             agents=["agent1"],
#             filter_fn=mock_functions["filter_fn"],
#             action_mapper_fn=mock_functions["action_mapper_fn"],
#             connector_fn=mock_functions["connector_fn"]
#         )
        
#         # Test when sim_results is 0 (success)
#         runner.sim_results = 0
#         assert runner.failed() is False
        
#         # Test when sim_results is not 0 (failure)
#         runner.sim_results = 1
#         assert runner.failed() is True

#     @patch('eprllib.Environment.EnvironmentRunner.api')
#     def test_make_eplus_args(self, mock_api, mock_env_config, mock_queues, mock_functions):
#         """Test the make_eplus_args method."""
#         # Create an instance of EnvironmentRunner
#         runner = EnvironmentRunner(
#             env_config=mock_env_config,
#             episode=0,
#             obs_queue=mock_queues["obs_queue"],
#             act_queue=mock_queues["act_queue"],
#             infos_queue=mock_queues["infos_queue"],
#             agents=["agent1"],
#             filter_fn=mock_functions["filter_fn"],
#             action_mapper_fn=mock_functions["action_mapper_fn"],
#             connector_fn=mock_functions["connector_fn"]
#         )
        
#         # Call make_eplus_args
#         args = runner.make_eplus_args()
        
#         # Check that the arguments are correct
#         assert "-w" in args
#         assert mock_env_config["epw_path"] in args
#         assert "-d" in args
#         assert mock_env_config["epjson_path"] in args
        
#         # Test with csv=True
#         runner.env_config["csv"] = True
#         args = runner.make_eplus_args()
#         assert "-r" in args

#     @patch('eprllib.Environment.EnvironmentRunner.api')
#     def test_progress_handler(self, mock_api, mock_env_config, mock_queues, mock_functions):
#         """Test the progress_handler method."""
#         # Create an instance of EnvironmentRunner
#         runner = EnvironmentRunner(
#             env_config=mock_env_config,
#             episode=0,
#             obs_queue=mock_queues["obs_queue"],
#             act_queue=mock_queues["act_queue"],
#             infos_queue=mock_queues["infos_queue"],
#             agents=["agent1"],
#             filter_fn=mock_functions["filter_fn"],
#             action_mapper_fn=mock_functions["action_mapper_fn"],
#             connector_fn=mock_functions["connector_fn"]
#         )
        
#         # Test with progress < 99
#         runner.progress_handler(50)
#         assert runner.is_last_timestep is False
        
#         # Test with progress >= 99
#         runner.progress_handler(99)
#         assert runner.is_last_timestep is True

#     @patch('eprllib.Environment.EnvironmentRunner.api')
#     def test_flush_queues(self, mock_api, mock_env_config, mock_queues, mock_functions):
#         """Test the _flush_queues method."""
#         # Create real queues for this test
#         obs_queue = Queue()
#         act_queue = Queue()
#         infos_queue = Queue()
        
#         # Add some items to the queues
#         obs_queue.put("obs")
#         act_queue.put("act")
#         infos_queue.put("info")
        
#         # Create an instance of EnvironmentRunner
#         runner = EnvironmentRunner(
#             env_config=mock_env_config,
#             episode=0,
#             obs_queue=obs_queue,
#             act_queue=act_queue,
#             infos_queue=infos_queue,
#             agents=["agent1"],
#             filter_fn=mock_functions["filter_fn"],
#             action_mapper_fn=mock_functions["action_mapper_fn"],
#             connector_fn=mock_functions["connector_fn"]
#         )
        
#         # Call _flush_queues
#         runner._flush_queues()
        
#         # Check that the queues are empty
#         assert obs_queue.empty()
#         assert act_queue.empty()
#         assert infos_queue.empty()



# 1. Import eprllib
from eprllib.Environment.EnvironmentConfig import EnvironmentConfig
from eprllib.Agents.AgentSpec import AgentSpec
from eprllib.Agents.ObservationSpec import ObservationSpec
from eprllib.Agents.ActionSpec import ActionSpec
from eprllib.Agents.Filters.FilterSpec import FilterSpec
from eprllib.Agents.ActionMappers.ActionMapperSpec import ActionMapperSpec
from eprllib.Agents.Rewards.RewardSpec import RewardSpec

# 2. Configure the Environment

environment = EnvironmentConfig()

# As in all EnergyPlus simulations, you need to provide the path to the EnergyPlus input files (IDF and EPW) and the output directory.
environment.generals(
    epjson_path = "path/to/your/model.epjson", 
    epw_path = "path/to/your/weather.epw", 
    output_path = "path/to/output/directory",
)

environment.connector(
    connector_fn: type[BaseConnector] = BaseConnector,
    connector_fn_config: Dict[str, Any] = {}
)

environment.agents(
    agents_config = {
        # This facilitate the configuration of the agent and provides a safety way of specifying the 
        # parameters of the agent. It is recommended to provides a intuitive name for the agent.
        "agent_1": AgentSpec(
            
            # The variables specified here will be the ones accesible from filter, action_mapper, and 
            # reward modules.
            observation = ObservationSpec(
                variables = [
                    ("variable_name_1", "variable_type_1"),
                    ("variable_name_2", "variable_type_2")
                ],
                internal_variables = [
                    ("variable_name_1", "variable_type_1"),
                    ("variable_name_2", "variable_type_2")
                ],
                meters = [
                    "meter_name_1",
                    "meter_name_2"
                ]
            ),
            
            # The action parameter is used to define the actuators that the agent can control in the 
            # EnergyPlus environment.
            action = ActionSpec(
                actuators = {
                    "actuator_name_1": ("actuator_key_1", "object_name_1", "field_name_1"),
                }
            ),
            
            # This object allows to transform the raw observations from EnergyPlus into a format that 
            # is more suitable for the RL algorithm. You can specify the type of filter to apply (e.g., 
            # normalization, scaling, etc.) and any relevant parameters for the chosen filter type.
            filter = FilterSpec(
                filter_fn = ,
                filter_fn_config = {}
            ),
            
            # The action mapper is responsible for mapping the actions chosen by the RL algorithm (the policy action) 
            # to the corresponding actuators in the EnergyPlus environment (actuator action). This module ensures that 
            # the actions taken by the agent are correctly translated into changes in the EnergyPlus simulation.
            action_mapper = ActionMapperSpec(
                
            ),
            
            # The reward module defines the reward function that the RL algorithm will use to evaluate the actions taken by the agent.
            reward = RewardSpec(
                
            )
        )
    }
)

# 3.  **Configure the RL Algorithm:** Set up an `RLlib algorithm <https://docs.ray.io/en/latest/rllib/rllib-algorithms.html>`_
#     to train the RL policy.
# 4.  **Execute Training:** Run the training process using `RLlib <https://docs.ray.io/en/latest/rllib/index.html>`_
#     or `Tune <https://docs.ray.io/en/latest/tune/index.html>`_.


from eprllib.Agents.AgentSpec import RewardSpec
from eprllib.Agents.Rewards.BaseReward import BaseReward
import pytest

class TestAgentspec:

    def test___getitem___1(self):
        """
        Test that __getitem__ method of RewardSpec correctly retrieves attribute values.
        """
        reward_fn = BaseReward
        reward_fn_config = {"test_key": "test_value"}
        reward_spec = RewardSpec(reward_fn=reward_fn, reward_fn_config=reward_fn_config)

        assert reward_spec["reward_fn"] == reward_fn
        assert reward_spec["reward_fn_config"] == reward_fn_config

    def test___getitem___non_existent_attribute(self):
        """
        Test that accessing a non-existent attribute via __getitem__ raises an AttributeError.
        """
        reward_spec = RewardSpec(BaseReward,{})
        with pytest.raises(AttributeError):
            reward_spec['non_existent_attribute']

    def test___init___1(self):
        """
        Test the initialization of RewardSpec with default parameters.
        """
        with pytest.raises(NotImplementedError) as e:
            RewardSpec().build()
        assert str(e.value) == "RewardSpec: No reward function provided."

    def test___setitem___1(self):
        """
        Test that the __setitem__ method correctly sets a new attribute on the RewardSpec object.
        """
        reward_spec = RewardSpec(BaseReward,{})
        key = 'new_attribute'
        with pytest.raises(KeyError)as e:
            reward_spec[key] = 'test_value'
            
        assert str(e.value) == f"'RewardSpec: Invalid key: {key}.'"

    def test_build_1(self):
        """
        Test that the build method of RewardSpec returns a dictionary containing
        the object's attributes when a valid reward function is provided.
        """
        class DummyReward(BaseReward):
            def __call__(self, env_object, infos):
                return 0.0

        reward_fn = DummyReward
        reward_fn_config = {"key": "value"}
        reward_spec = RewardSpec(reward_fn=reward_fn, reward_fn_config=reward_fn_config)

        result = reward_spec.build()

        assert isinstance(result, dict)
        assert result['reward_fn'] == reward_fn
        assert result['reward_fn_config'] == reward_fn_config

    # def test_reward_spec_init_with_invalid_reward_fn(self):
    #     """
    #     Test initializing RewardSpec with an invalid reward_fn that is not a BaseReward instance.
    #     This should raise a ValueError as per the validation in the RewardSpec.validation_rew_config method.
    #     """
    #     with pytest.raises(TypeError) as e:
    #         RewardSpec(reward_fn="invalid_reward_fn", reward_fn_config={}).build()
    #     assert str(e.value) == "issubclass() arg 1 must be a class"

    # def test_reward_spec_init_with_not_implemented_reward_fn(self):
    #     """
    #     Test initializing RewardSpec with NotImplemented as the reward_fn.
    #     This should raise a NotImplementedError as per the validation in the RewardSpec.validation_rew_config method.
    #     """
    #     with pytest.raises(NotImplementedError) as e:
    #         RewardSpec(reward_fn=NotImplemented).build()
            
    #     assert str(e.value) == "No reward function provided."



"""
Specification for the filter functions
===========================================
This module defines the ``FilterSpec`` class, which is used to specify the 
configuration of filter functions for agents in reinforcement learning environments.
It ensures that the filter function is properly defined and adheres to the expected interface.
"""
from typing import Dict, Any, Optional, Type

from eprllib.Agents.Filters.BaseFilter import BaseFilter
from eprllib import logger

class FilterSpec:
    """
    FilterSpec is the base class for a filter specification to safe configuration of the object.
    """
    filter_fn: Optional[Type[BaseFilter]] = None
    filter_fn_config: Dict[str, Any] = {}
    
    def __init__(
        self,
        filter_fn: Optional[Type[BaseFilter]] = None,
        filter_fn_config: Dict[str, Any] = {}
    ):
        """
        Construction method.
        
        Args:
            filter_fn (BaseFilter): The filter function takes the arguments agent_id, observation and returns the
            observation filtered. See ``eprllib.Agents.Filters`` for examples.
            
            filter_fn_config (Dict[str, Any]): The configuration of the filter function.
        """
        self.filter_fn = filter_fn
        self.filter_fn_config = filter_fn_config
    
    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)

    def __setitem__(self, key: str, value: Any) -> None:
        valid_keys = self.__dict__.keys()
        if key not in valid_keys:
            msg = f"FilterSpec: Invalid key: {key}."
            logger.error(msg)
            raise KeyError(msg)
        setattr(self, key, value)
        
    def build(self) -> Dict[str, Any]:
        """
        This method is used to build the FilterSpec object.
        """
        if self.filter_fn is None:
            msg = "FilterSpec: No filter function provided."
            logger.error(msg)
            raise ValueError(msg)
        
        return vars(self)
    


"""
EnvironmentConfig Utilities
==============================

Work in progress...
"""

from typing import Set, Dict, Optional, List, Any
import inspect
from eprllib.Environment.EnvironmentConfig import EnvironmentConfig
from gymnasium.spaces import Box, Discrete
import sys
import numpy as np
from typing import get_origin, get_args, Union, Tuple

from eprllib import logger


def env_config_validation(MyEnvConfig: EnvironmentConfig) -> bool:
    """
    Validate the EnvConfig object before to be used in the env_config parameter of RLlib environment config.
    """
    # Check that the variables defined in EnvConfig are the allowed in the EnvConfig base
    # class.
    allowed_vars = inspect.get_annotations(EnvironmentConfig).keys()
    for var in vars(MyEnvConfig):
        if var not in allowed_vars:
            msg = f"EnvConfigUtils: The variable '{var}' is not allowed in EnvConfig. Allowed variables are: {allowed_vars}"
            logger.error(msg)
            raise ValueError(msg)
    return True

def to_json(
    MyEnvConfig: EnvironmentConfig,
    output_path: Optional[str] = None
    ) -> str:
    """Convert an EnvConfig object into a json string before to be used in the env_config parameter of RLlib environment config.

    Args:
        MyEnvConfig (EnvConfig): _description_
        output_path (str, optional): _description_. Defaults to None.

    Returns:
        str: _description_
    """
    import json
    import time
    
    env_config_json = json.dumps(MyEnvConfig.to_dict())

    # generate a unique number based on time
    time_id = str(int(time.time()))
    # check the implementation of output_path
    if output_path is None:
        output_path = './'
    path = output_path+f'/{time_id}_env_config.json'
    # save the json string to a file
    with open(path, 'x') as f:
        f.write(env_config_json)
    
    logger.info(f"EnvConfigUtils: EnvConfig saved to {path}")
    
    return path

def from_json(
    path: str
    ) -> EnvironmentConfig:
    """Convert a json file into an EnvConfig object before to be used in the env_config parameter of RLlib environment config.

    Args:
        path (str): _description_

    Returns:
        EnvConfig: _description_
    """
    import json
    with open(path, 'r') as f:
        env_config_json = f.read()
    env_config_dict = json.loads(env_config_json)
    env_config = EnvironmentConfig(**env_config_dict)
    return env_config

def continuous_action_space():
    """
    This method construct the action space of the environment.
    
    Returns:
        gym.Box: Continuous action space with limits between [0,1].
    """
    return Box(low=0.0, high=1.0, shape=(1,), dtype=np.float64)

def discrete_action_space(n:int=2):
    """
    This method construct the action space of the environment.
    
    Returns:
        gym.Discrete: Discrete action space with limits between [0,10] and a step of 1.
    """
    return Discrete(n)


def variable_checking(
    epJSON_file:str,
) -> Set[Any]|None:
    """
    This function check if the epJSON file has the required variables.

    Args:
        epJSON_file(str): path to the epJSON file.

    Return:
        set: list of missing variables.
    """
    pass

def validate_properties(obj: Any, expected_types: Dict[str, Any]) -> Tuple[bool, List[str]]:
    """
    Enhanced version that supports Union types and optional properties
    
    Args:
        obj: The object to validate
        expected_types: A dictionary mapping property names to either:
                       - A type
                       - A tuple of types (Union)
                       - (type, bool) tuple where bool indicates if property is optional
    """
    errors: List[str] = []
    is_valid = True
    
    for prop_name, type_spec in expected_types.items():
        # Handle optional properties
        is_optional = False
        expected_type = type_spec
        
        if isinstance(type_spec, tuple):
            if len(type_spec) == 2:
                if isinstance(type_spec[1], bool):
                    expected_type, is_optional = type_spec
            
        # Check if property exists
        if not hasattr(obj, prop_name):
            if not is_optional:
                errors.append(f"Missing required property: {prop_name}")
                is_valid = False
            continue
            
        actual_value = getattr(obj, prop_name)
        
        # Handle None for optional properties
        if actual_value is None and is_optional:
            continue
            
        # Handle union types
        valid_types = (expected_type,) if isinstance(expected_type, type) else expected_type
        
        def is_instance_of_type(value, types):
            if isinstance(types, _SpecialGenericAlias):
                types = (types,)
            for typ in types:
                origin = get_origin(typ)
                if origin:
                    if origin is Union:
                        if any(is_instance_of_type(value, get_args(typ))):
                            return True
                    elif isinstance(value, origin):
                        return True
                elif isinstance(value, typ):
                    return True
            return False

        if not is_instance_of_type(actual_value, valid_types):
            errors.append(
                f"Property '{prop_name}' has incorrect type. "
                f"Expected {', '.join(t.__name__ if isinstance(t, type) else str(t) for t in valid_types)}, "
                f"got {type(actual_value).__name__}"
            )
            is_valid = False
    
    return is_valid, errors



"""
Episode Functions Base Class
=============================

This module contains the base class for the ``Episode`` functions. This class is used to define the ``Episode`` configuration
for the EnergyPlus environment.

The methods provided here are used during inizialization and execution of the environment.
You can overwrite the following methods for sophisticated executions:

    - ``setup(self)``
    - ``get_episode_config(self, env_config: Dict[str, Any])``
    - ``get_episode_agents(self, env_config: Dict[str, Any], possible_agents: List[str])``
    - ``get_timestep_agents(self, env_config: Dict[str, Any], possible_agents: List[str])``

The ``DefaultEpisode`` configuration will be used by default.
"""

from typing import Dict, Any, List

from eprllib import logger

class BaseEpisode:
    """
    This class contains the methods to configure the episode in EnergyPlus with RLlib.
    """
    episode_fn_config: Dict[str, Any] = {}
    
    def __init__(
        self,
        episode_fn_config: Dict[str, Any] = {}
    ):
        """
        Initializes the BaseEpisode object.

        Args:
            episode_fn_config (Dict[str, Any]): Configuration dictionary for the episode function.
        """
        # Set the episode_fn_config attribute
        self.episode_fn_config = episode_fn_config
        
        logger.info(f"BaseEpisode: The BaseEpisode was correctly inicializated with {self.episode_fn_config} config.")
        
        # Make sure, `setup()` is only called once, no matter what.
        if hasattr(self, "_is_setup") and self._is_setup:
            raise RuntimeError(
                "``BaseActionMapper.setup()`` called twice within your ActionMapper implementation "
                f"{self}! Make sure you are using the proper inheritance order "
                " and that you are NOT overriding the constructor, but "
                "only the ``setup()`` method of your subclass."
            )
        try:
            self.setup()
        except AttributeError as e:
            raise e

        self._is_setup:bool = True
    
    # ===========================
    # === OVERRIDABLE METHODS ===
    # ===========================
    
    def setup(self) -> None:
        """
        Sets up the episode function.
        
        This is called automatically during the __init__ method of this class.
        """
        pass
    
    
    def get_episode_config(self, env_config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Returns the episode configuration for the EnergyPlus environment.

        Args:
            env_config (Dict[str, Any]): The environment configuration.

        Returns:
            Dict[str, Any]: The episode configuration.
        """
        return env_config
    
    
    def get_episode_agents(self, possible_agents: List[str]) -> List[str]:
        """
        Returns the agents for the episode configuration in the EnergyPlus environment.

        Args:
            env_config (Dict[str, Any]): The environment configuration.
            possible_agents (List[str]): List of possible agents.

        Returns:
            Dict[str, Any]: The agents that are acting for the episode configuration. Default: possible_agents list.
        """
        return possible_agents
    
    
    def get_timestep_agents(self, possible_agents: List[str]) -> List[str]:
        """
        Returns the agents for the timestep configuration in the EnergyPlus environment.

        Args:
            env_config (Dict[str, Any]): The environment configuration.
            possible_agents (List[str]): List of possible agents.

        Returns:
            Dict[str, Any]: The agents that are acting for the timestep configuration. Default: possible_agents list.
        """
        return possible_agents


from eprllib.Agents.ActionSpec import ActionSpec
import io
import pytest
import sys

class TestActionspec:

    def test___getitem___1(self):
        """
        Test the __getitem__ method of ActionSpec class.

        This test verifies that the __getitem__ method correctly retrieves
        the value of an existing attribute using the key.
        """
        # Create an instance of ActionSpec with some actuators
        action_spec = ActionSpec(actuators={'my_actuator': ('type1', 'component1', 'control1')})

        # Test retrieving the 'actuators' attribute
        assert action_spec['actuators'] == {'my_actuator': ('type1', 'component1', 'control1')}

    def test___getitem___invalid_key(self):
        """
        Test __getitem__ method with an invalid key.
        This tests the edge case where a non-existent attribute is accessed,
        which should raise an AttributeError according to the method's implementation.
        """
        action_spec = ActionSpec()
        with pytest.raises(AttributeError):
            _ = action_spec["non_existent_key"]

    def test___init___2(self):
        """
        Test the initialization of ActionSpec with non-None actuators.

        This test verifies that when actuators are provided during initialization,
        they are correctly assigned to the ActionSpec instance.
        """
        actuators = {'actuator_1': ("type1", "component1", "control1"),'actuator_2': ("type2", "component2", "control2")}
        action_spec = ActionSpec(actuators=actuators)
        assert action_spec.actuators == actuators

    def test___init___invalid_actuator_length(self):
        """
        Test the __init__ method with an invalid actuator tuple length.
        This test verifies that a ValueError is raised when an actuator tuple
        does not contain exactly 3 elements.
        """
        with pytest.raises(ValueError) as excinfo:
            ActionSpec(actuators={"invalid_actuator": ("invalid", "tuple")}).build()
        assert "ActionSpec: The actuators must be defined as a dict with actuator names as keys and tuples of 3 elements" in str(excinfo.value)

    def test___init___invalid_actuator_type(self):
        """
        Test the __init__ method with an invalid actuator type.
        This test verifies that a ValueError is raised when the actuators
        argument is not a list of tuples.
        """
        with pytest.raises(ValueError) as excinfo:
            ActionSpec(actuators=["invalid_actuator"]).build()
        assert "ActionSpec: The actuators must be defined as a dict with actuator names as keys and tuples of 3 elements" in str(excinfo.value)

    def test___setitem___2(self):
        """
        Test that __setitem__ successfully sets a valid key-value pair.

        This test verifies that when a valid key is provided to __setitem__,
        the corresponding attribute is set to the given value without raising
        any exceptions.
        """
        action_spec = ActionSpec()
        action_spec['actuators'] = {'my_actuator':('type1', 'component1', 'control1')}
        assert action_spec.actuators == {'my_actuator':('type1', 'component1', 'control1')}

    def test___setitem___invalid_key(self):
        """
        Test that __setitem__ raises a KeyError when an invalid key is provided.
        This tests the explicit error handling in the method for invalid keys.
        """
        action_spec = ActionSpec()
        with pytest.raises(KeyError, match="Invalid key: invalid_key"):
            action_spec["invalid_key"] = "some_value"

    def test___setitem___invalid_key_2(self):
        """
        Tests the __setitem__ method of ActionSpec when an invalid key is provided.
        This test verifies that a KeyError is raised when attempting to set an item
        with a key that is not present in the object's __dict__.
        """
        action_spec = ActionSpec()
        with pytest.raises(KeyError) as exc_info:
            action_spec["invalid_key"] = "some_value"
        assert str(exc_info.value) == "'ActionSpec: Invalid key: invalid_key.'"

    # def test_build_1(self):
    #     """
    #     Test that the build method raises a ValueError when actuators is not a list.
    #     """
    #     action_spec = ActionSpec(actuators="not a list")
    #     with pytest.raises(ValueError) as context:
    #         action_spec.build()

    #     assert "The actuators must be defined as a list of tuples but <class 'str'> was given." == str(context.value)

    # def test_build_2(self):
    #     """
    #     Test that build() raises ValueError when actuators contain a non-tuple element.

    #     This test verifies that the build() method raises a ValueError when the
    #     actuators list contains an element that is not a tuple. It covers the path
    #     where isinstance(self.actuators, list) is True, but isinstance(actuator, tuple)
    #     is False for at least one actuator.
    #     """
    #     action_spec = ActionSpec(actuators=[('a', 'b', 'c'), 'not_a_tuple'])
    #     with pytest.raises(ValueError, match="The actuators must be defined as a list of tuples but <class 'str'> was given."):
    #         action_spec.build()

    # def test_build_invalid_actuator_element_type(self):
    #     """
    #     Test the build method with an invalid actuator element type (not a tuple).
    #     This should raise a ValueError.
    #     """
    #     action_spec = ActionSpec(actuators=[["invalid"]])
    #     with pytest.raises(ValueError) as exc_info:
    #         action_spec.build()
    #     assert str(exc_info.value) == "The actuators must be defined as a list of tuples but <class 'list'> was given."

    def test_build_invalid_actuator_tuple_length(self):
        """
        Test the build method with an invalid actuator tuple length (not 3 elements).
        This should raise a ValueError.
        """
        action_spec = ActionSpec(actuators=[("invalid", "tuple")])
        with pytest.raises(ValueError) as excinfo:
            action_spec.build()
        assert "ActionSpec: The actuators must be defined as a dict with actuator names as keys and tuples of 3 elements"  in str(excinfo.value)

    # def test_build_invalid_actuator_type(self):
    #     """
    #     Test the build method with an invalid actuator type (not a list).
    #     This should raise a ValueError.
    #     """
    #     action_spec = ActionSpec(actuators="invalid")
    #     with pytest.raises(ValueError) as exc_info:
    #         action_spec.build()
    #     assert str(exc_info.value) == "The actuators must be defined as a list of tuples but <class 'str'> was given."



"""
Specification for the observation space and parameters
===========================================================
This module defines the ``ObservationSpec`` class, which is used to specify the configuration of 
observation space and parameters for agents in reinforcement learning environments.

It ensures that the observation space is properly defined and adheres to the expected interface.
"""
from typing import Dict, List, Tuple, Optional, Any

from eprllib.Agents import (
    SIMULATION_PARAMETERS,
    ZONE_SIMULATION_PARAMETERS, 
    PREDICTION_VARIABLES,
    PREDICTION_HOURS,
    VALID_USER_TYPES,
    VALID_ZONE_TYPES
)
from eprllib import logger

class ObservationSpec:
    """
    ObservationSpec is the base class for an observation specification to safe configuration of the object.
    """
    variables: Optional[List[Tuple[str, str]]] = None
    internal_variables: Optional[List[Tuple[str, str]]] = None
    meters: Optional[List[str]] = None
    simulation_parameters: Dict[str, bool] = {}
    zone_simulation_parameters: Dict[str, bool] = {}
    use_one_day_weather_prediction: bool = False
    weather_prediction_hours: int = PREDICTION_HOURS
    prediction_variables: Dict[str, bool] = {}
    use_actuator_state: bool = False
    other_obs: Dict[str, float | int] = {}
    user_occupation_function: bool = False
    user_occupation_forecast: bool = False
    user_type: str = VALID_USER_TYPES[0]
    zone_type: str = VALID_ZONE_TYPES[0]
    occupation_schedule: Optional[Tuple[str, str, str]] = None
    occupation_prediction_hours: int = 24
    confidence_level: float = 0.95
    lambdaa: float = 0.05
    
    def __init__(
        self,
        variables: Optional[List[Tuple[str, str]]] = None,
        internal_variables: Optional[List[Tuple[str, str]]] = None,
        meters: Optional[List[str]] = None,
        simulation_parameters: Dict[str, bool] = {},
        zone_simulation_parameters: Dict[str, bool] = {},
        use_one_day_weather_prediction: bool = False,
        weather_prediction_hours: int = PREDICTION_HOURS,
        prediction_variables: Dict[str, bool] = {},
        use_actuator_state: bool = False,
        other_obs: Dict[str, float | int] = {},
        user_occupation_function: bool = False,
        user_occupation_forecast: bool = False,
        user_type: str = VALID_USER_TYPES[0],
        zone_type: str = VALID_ZONE_TYPES[0],
        occupation_schedule: Optional[Tuple[str, str, str]] = None,
        occupation_prediction_hours: int = 24,
        confidence_level: float = 0.95,
        lambdaa: float = 0.05
        ) -> None:
        """
        Construction method.
        
        Args:
        
            variables (List[Tuple[str, str]]): Variables represent time series output variables in the simulation. There are thousands
                of variables made available based on the specific configuration. A user typically requests
                variables to be in their output files by adding Output:Variable objects to the input file. It
                is important to note that if the user does not request these variables, they are not tracked,
                and thus not available on the API.
            internal_variables (List[str]): Internal variables form a category of built-in data accessible for eprllib.
                They are internal in that they access information about the input file from inside EnergyPlus. Internal variables 
                are automatically made available. The EDD file lists the specific internal variable types, their unique
                identifying names, and the units. The rest of this section provides information about specific internal variables.
                To see the EDD file, run the simulation to generate the output files and configure an object in Output:EnergyManagementSystem
                with "Internal Variable Availability Dictionary Reporting" to Verbose.
            meters (List[str]): Meters represent groups of variables which are collected together, much like a meter on
                a building which represents multiple energy sources. Meters are handled the same way as
                variables, except that meters do not need to be requested prior running a simulation. From
                an API standpoint, a client must simply get a handle to a meter by name, and then access
                the meter value by using a get-value function on the API.
            simulation_parameters (Dict[str, bool]): A number of parameters are made available as they vary through the
                simulation, including the current simulation day of week, day of year, hour, and many other
                things. These do not require a handle, but are available through direct function calls. The keys (all by default False) are:  
                    
                    'actual_date_time', 'actual_time', 'current_time', 'day_of_month', 'day_of_week', 'day_of_year', 'holiday_index',
                    'hour', 'minutes', 'month', 'num_time_steps_in_hour', 'year', 'is_raining', 'sun_is_up', 'today_weather_albedo_at_time',
                    'today_weather_beam_solar_at_time', 'today_weather_diffuse_solar_at_time', 'today_weather_horizontal_ir_at_time',
                    'today_weather_is_raining_at_time', 'today_weather_is_snowing_at_time', 'today_weather_liquid_precipitation_at_time',
                    'today_weather_outdoor_barometric_pressure_at_time', 'today_weather_outdoor_dew_point_at_time', 
                    'today_weather_outdoor_dry_bulb_at_time', 'today_weather_outdoor_relative_humidity_at_time', 'today_weather_sky_temperature_at_time',
                    'today_weather_wind_direction_at_time', 'today_weather_wind_speed_at_time', 'tomorrow_weather_albedo_at_time',
                    'tomorrow_weather_beam_solar_at_time', 'tomorrow_weather_diffuse_solar_at_time', 'tomorrow_weather_horizontal_ir_at_time',
                    'tomorrow_weather_is_raining_at_time', 'tomorrow_weather_is_snowing_at_time', 'tomorrow_weather_liquid_precipitation_at_time',
                    'tomorrow_weather_outdoor_barometric_pressure_at_time', 'tomorrow_weather_outdoor_dew_point_at_time', 'tomorrow_weather_outdoor_dry_bulb_at_time',
                    'tomorrow_weather_outdoor_relative_humidity_at_time', 'tomorrow_weather_sky_temperature_at_time', 'tomorrow_weather_wind_direction_at_time',
                    'tomorrow_weather_wind_speed_at_time'
            
            zone_simulation_parameters (Dict[str, bool]):  A number of parameters are made available as they vary through the
                simulation for thermal zones. The keys (all by default False) are:
                
                    'system_time_step', 'zone_time_step', 'zone_time_step_number'
                
            use_one_day_weather_prediction (bool): We use the internal variables of EnergyPlus to provide with a 
            prediction of the weather time ahead. You can specify the `weather_prediction_hours` and the prediction variables
            listed on `prediction_variables`.
            weather_prediction_hours (int): Default is 24
            prediction_variables (Dict[str, bool]): See `use_one_day_weather_prediction`. The variables are (by default 
                all False):
            
                    'albedo', 'beam_solar', 'diffuse_solar', 'horizontal_ir', 'is_raining', 'is_snowing', 'liquid_precipitation',
                    'outdoor_barometric_pressure', 'outdoor_dew_point', 'outdoor_dry_bulb', 'outdoor_relative_humidity',
                    'sky_temperature', 'wind_direction', 'wind_speed'.
                
            use_actuator_state (bool): Define if the actuator state will be used as an observation for the agent.
            other_obs (Dict[str, float | int]): Custom observation dictionary.
            user_occupation_function (bool): Define if the user occupation function will be used. Default is False.
            user_occupation_forecast (bool): Define if the user occupation forecast will be used. Default is False.
            user_type (str): Type of user. Default is 'Residential'
            zone_type (str): Type of zone. Default is 'Office'
            occupation_schedule (Tuple[str, str, str]): Occupation schedule for the user occupation forecast. Default is None.
            occupation_prediction_hours (int): Occupation prediction hours for the user occupation forecast. Default is 24.
            confidence_level (float): Confidence level for the user occupation forecast. Default is 0.95.
            lambdaa (float): Decay value of the confidence level. Default is 0.05.

        Raises:
        
            ValueError: If the user type or zone type is not valid, if the weather prediction hours or occupation 
                prediction hours are not between 1 and 24, if the occupation schedule is not provided when user 
                occupation forecast is True, or if no variable/meter/actuator/parameter is defined in the observation.
            KeyError: If an invalid key is provided when setting an item.
            
        """
        # Variables
        self.variables = variables
        # Internal variables
        self.internal_variables = internal_variables
        # Meters
        self.meters = meters
        # Simulation parameters
        self.simulation_parameters = SIMULATION_PARAMETERS.copy()
        
        # Update the boolean values in the self.simulation_parameters Dict.
        self.simulation_parameters.update(simulation_parameters)
        
        # Zone simulation parameters.
        self.zone_simulation_parameters = ZONE_SIMULATION_PARAMETERS.copy()
        
        # Update the boolean values in the self.zone_simulation_parameters Dict.
        self.zone_simulation_parameters.update(zone_simulation_parameters)
        
        # Prediction weather.
        self.prediction_variables = PREDICTION_VARIABLES.copy()
        self.use_one_day_weather_prediction = use_one_day_weather_prediction
        if self.use_one_day_weather_prediction:
            # Update the boolean values in the self.prediction_variables Dict.
            self.prediction_variables.update(prediction_variables)
        self.weather_prediction_hours = weather_prediction_hours
            
        # Actuator value in the observation.
        self.use_actuator_state = use_actuator_state
        
        # Custom observation dict.
        self.other_obs = other_obs
        
        # User occupation forecast profile.
        self.user_occupation_forecast = user_occupation_forecast
        self.user_occupation_function = user_occupation_function
        if self.user_occupation_forecast: # This ensure that if forecast is used, the occupancy function is used as well.
            self.user_occupation_function = True
        
        # User type and zone type.
        self.user_type = user_type
        
        if user_type not in VALID_USER_TYPES:
            msg = f"ObservationSpec: User type '{user_type}' is not valid. Options: {VALID_USER_TYPES}"
            logger.error(msg)
            raise ValueError(msg)
        
        self.zone_type = zone_type
        if zone_type not in VALID_ZONE_TYPES:
            msg = f"ObservationSpec: Zone type '{zone_type}' is not valid. Options: {VALID_ZONE_TYPES}"
            logger.error(msg)
            raise ValueError(msg)
        
        self.confidence_level = confidence_level
        
        # Occupation forecast.
        self.occupation_schedule = occupation_schedule
        self.occupation_prediction_hours = occupation_prediction_hours
        self.lambdaa = lambdaa
        
    def __getitem__(self, key:str) -> Any:
        return getattr(self, key)

    def __setitem__(self, key:str, value:Any) -> None:
        valid_keys = self.__dict__.keys()
        if key not in valid_keys:
            msg = f"ObservationSpec: Invalid key: {key}."
            logger.error(msg)
            raise KeyError(msg)
        setattr(self, key, value)
        
    
    def build(self) -> Dict[str, Any]:
        """
        This method is used to build the ObservationSpec object.
        """
        # Check that the keys introduced in the Dict are admissible values.
        admissible_values = list(SIMULATION_PARAMETERS.keys())
        for key in self.simulation_parameters.keys():
            if key not in admissible_values:
                msg = f"ObservationSpec: The key '{key}' is not admissible in the simulation_parameters. The admissible values are: {admissible_values}"
                logger.error(msg)
                raise ValueError(msg)
        
        # Check that the keys introduced in the Dict are admissible values.
        admissible_values = list(ZONE_SIMULATION_PARAMETERS.keys())
        for key in self.zone_simulation_parameters.keys():
            if key not in admissible_values:
                msg = f"ObservationSpec: The key '{key}' is not admissible in the zone_simulation_parameters. The admissible values are: {admissible_values}"
                logger.error(msg)
                raise ValueError(msg)
        
        if self.use_one_day_weather_prediction:
            admissible_values = list(PREDICTION_VARIABLES.keys())
            for key in self.prediction_variables.keys():
                if key not in admissible_values:
                    msg = f"ObservationSpec: The key '{key}' is not admissible in the prediction_variables. The admissible values are: {admissible_values}"
                    logger.error(msg)
                    raise ValueError(msg)
        
        if self.weather_prediction_hours <= 0 or self.weather_prediction_hours > 24:
            self.weather_prediction_hours = PREDICTION_HOURS
            logger.warning(f"ObservationSpec: The variable 'weather_prediction_hours' must be between 1 and 24. It is taken the value of {self.weather_prediction_hours}. The value of 24 is used.")
        
        if self.occupation_prediction_hours <= 0 or self.occupation_prediction_hours > 24:
                self.occupation_prediction_hours = 24
                logger.warning(f"ObservationSpec: The variable 'occupation_prediction_hours' must be between 1 and 24. It is taken the value of {self.occupation_prediction_hours}. The value of 24 is used.")
        
        if self.user_occupation_forecast:
            if self.occupation_schedule is None:
                msg = "ObservationSpec: occupation_schedule must be provided if user_occupation_forecast is True."
                logger.error(msg)
                raise ValueError(msg)
            else:
                assert isinstance(self.occupation_schedule, tuple), "occupation_schedule must be a tuple."
        # Check that at least one variable/meter/actuator/parameter is defined.
        counter = 0
        if self.variables is not None:
            counter += len(self.variables)
        if self.internal_variables is not None:
            counter += len(self.internal_variables)
        if self.meters is not None:
            counter += len(self.meters)
        counter += sum([1 for value in self.simulation_parameters.values() if value])
        counter += sum([1 for value in self.zone_simulation_parameters.values() if value])
        counter += sum([1 for value in self.prediction_variables.values() if value])
        if self.use_actuator_state:
            counter += 1
        counter += len(self.other_obs)
        
        if counter == 0:
            msg = "ObservationSpec: At least one variable/meter/actuator/parameter must be defined in the observation."
            logger.error(msg)
            raise ValueError(msg)
        
        return vars(self)
            


"""
RUN DRL CONTROLS
=================

This script execute the conventional controls in the evaluation scenario.
"""
from typing import Dict, Any, List
from ray.rllib.policy.policy import Policy
from pathlib import Path
from ray.rllib.core.rl_module.rl_module import RLModule
from eprllib.Environment.Environment import Environment
import torch
import numpy as np
import tree

def generate_experience(
    env: Environment,
    env_config: Dict[str, Any],
    policies: Dict[str, Policy],
    num_episodes: int = 1
) -> Dict[str, Any]:
    """
    Function to generate experience and store it in a structured dictionary.

    Args:
        env (Environment): The environment class to instantiate.
        env_config (Dict[str, Any]): Configuration for the environment.
        policies (Dict[str, Policy]): A dictionary mapping agent IDs to their policies.
        num_episodes (int, optional): The number of episodes to simulate. Defaults to 1.

    Returns:
        Dict[str, Any]: A dictionary containing the generated experience
                        structured as:
                        {
                            'experiment': {
                                'episode_1': {
                                    'timestep_1': {
                                        'agent_name_1': {
                                            'observation': value,
                                            'action': value,
                                            'reward': value,
                                            'done': value,
                                            'info': value,
                                            'state': value (if rnn_use[agent] is True)
                                        },
                                        'agent_name_2': { ... }
                                    },
                                    'timestep_2': { ... }
                                },
                                'episode_2': { ... }
                            }
                        }
    """
    # Instantiate the environment
    env_instance: Environment = env(env_config)
    
    # Main dictionary to store all experiment data
    experiment_data: Dict[str, Any] = {'experiment': {}}
    
    for episode_idx in range(1, num_episodes + 1):
        
        # Dictionary to store data for the current episode
        episode_data: Dict[str, Any] = {}
        timestep_counter = 0
        
        # Initialize RNN states for agents that use them
        agent_list = policies.keys()
        current_states = {
            agent: policies[agent].get_initial_state()
            for agent in agent_list if policies[agent].is_recurrent()# Check rnn_use for each agent
        }
        
        # Reset the environment and get the initial observation and infos.
        obs, info = env_instance.reset()

        # List the agents in the environment for the first timestep.
        _agent_ids: List[str] = list(obs.keys())

        # Reward for timestep 0 (initial state, no action taken yet).
        reward = {agent: 0 for agent in _agent_ids}
        # Control variable to finish the episode.
        done = {"__all__": False}
        truncateds = {agent: False for agent in _agent_ids} # Initialize truncateds for consistency

        # --- Store the experience for the initial timestep (timestep_0 or timestep_1 depending on convention) ---
        timestep_counter += 1
        timestep_data: Dict[str, Any] = {}
        for agent in _agent_ids:
            agent_experience: Dict[str, Any] = {
                'observation': obs[agent],
                'action': None, # No action taken yet at timestep 0
                'reward': reward[agent],
                'done': done["__all__"],
                'truncated': truncateds[agent], # Add truncateds
                'info': info[agent]
            }
            if policies[agent].is_recurrent(): # Check rnn_use for the specific agent
                agent_experience['state'] = current_states.get(agent) # Use .get() in case agent not in current_states
            timestep_data[agent] = agent_experience
        episode_data[f'timestep_{timestep_counter}'] = timestep_data
        
        # Loop to generate the experience for subsequent timesteps.
        while not done["__all__"]:
            # Action dictionary for the current timestep.
            action: Dict[str, Any] = {}
            new_states = {}
            
            # Get the action for each agent in the timestep.
            for agent in _agent_ids:
                if policies[agent].is_recurrent(): # Check rnn_use for the specific agent
                    # Compute action with RNN state
                    agent_action, agent_state, _ = policies[agent].compute_single_action(
                        obs[agent], state=current_states.get(agent), explore=False
                    )
                    action[agent] = agent_action
                    new_states[agent] = agent_state
                    
                    # Update current_states for the next iteration if RNN is used by any agent
                    current_states[agent] = new_states[agent]
                    
                else:
                    # Compute action without RNN state
                    agent_action, _, _ = policies[agent].compute_single_action(
                        obs[agent], explore=False
                    )
                    action[agent] = agent_action

                # assert type(action[agent]) is float|int, f"Action must be a float or int, but got {type(action[agent])} for agent {agent} at timestep {timestep_counter}."
            # Execute the action in the environment.
            new_obs, new_reward, done, truncateds, new_info = env_instance.step(action)

            # Update the observation, reward, done, truncateds, and info for the next timestep.
            obs = new_obs
            info = new_info
            reward = new_reward
            
            # List the agents in the environment for the next timestep.
            # This handles cases where agents might leave the environment.
            _agent_ids = list(obs.keys())
            
            # --- Store the experience for the current timestep ---
            timestep_counter += 1
            timestep_data = {}
            for agent in _agent_ids:
                agent_experience = {
                    'observation': obs[agent],
                    'action': action.get(agent), # Action taken to reach this state
                    'reward': reward[agent],
                    'done': done["__all__"],
                    'truncated': truncateds["__all__"],
                    'info': info[agent]
                }
                if policies[agent].is_recurrent(): # Check rnn_use for the specific agent
                    # Store the state *after* computing the action for this timestep
                    agent_experience['state'] = current_states.get(agent) # Use .get()
                timestep_data[agent] = agent_experience
            episode_data[f'timestep_{timestep_counter}'] = timestep_data

        # Add the completed episode data to the experiment data
        experiment_data['experiment'][f'episode_{episode_idx}'] = episode_data

    return experiment_data

def generate_experience_V2(
    env: Environment,
    env_config: Dict[str, Any],
    rl_module_checpoints: Dict[str, str],
    policy_names: Dict[str, str],
    num_episodes: int = 1
) -> Dict[str, Any]:
    """
    Function to generate experience and store it in a structured dictionary.

    Args:
        env (Environment): The environment class to instantiate.
        env_config (Dict[str, Any]): Configuration for the environment.
        policies (Dict[str, Policy]): A dictionary mapping agent IDs to their policies.
        num_episodes (int, optional): The number of episodes to simulate. Defaults to 1.

    Returns:
        Dict[str, Any]: A dictionary containing the generated experience
                        structured as:
                        {
                            'experiment': {
                                'episode_1': {
                                    'timestep_1': {
                                        'agent_name_1': {
                                            'observation': value,
                                            'action': value,
                                            'reward': value,
                                            'done': value,
                                            'info': value,
                                            'state': value (if rnn_use[agent] is True)
                                        },
                                        'agent_name_2': { ... }
                                    },
                                    'timestep_2': { ... }
                                },
                                'episode_2': { ... }
                            }
                        }
    """
    # Instantiate the environment
    env_instance: Environment = env(env_config)
    
    # Main dictionary to store all experiment data
    experiment_data: Dict[str, Any] = {'experiment': {}}
    
    agent_list: List[str] = list(rl_module_checpoints.keys())
    rl_module_dict: Dict[str, RLModule] = {}
    for agent in agent_list:
        rl_module_dict[agent] = RLModule.from_checkpoint(Path(rl_module_checpoints[agent]) / "learner_group" / "learner" / "rl_module" / policy_names[agent])

        # assert type(rl_module_dict[agent]) == RLModule, f"assert rl_module_dict[{agent}] is not RLModule, is {type(rl_module_dict[agent])}."
    
    for episode_idx in range(1, num_episodes + 1):
        
        # Dictionary to store data for the current episode
        episode_data: Dict[str, Any] = {}
        timestep_counter = 0
        prev_action: Dict[str, Any] = {agent: 0 for agent in agent_list}
        
        # Initialize RNN states for agents that use them
        current_states: Dict[str, Any]= {
            agent: rl_module_dict[agent].get_initial_state()
            for agent in agent_list if rl_module_dict[agent].is_stateful()# Check rnn_use for each agent
        }
        
        # Reset the environment and get the initial observation and infos.
        obs, info = env_instance.reset()

        # List the agents in the environment for the first timestep.
        _agent_ids: List[str] = list(obs.keys())

        # Reward for timestep 0 (initial state, no action taken yet).
        reward = {agent: 0 for agent in _agent_ids}
        # Control variable to finish the episode.
        done = {"__all__": False}
        truncateds = {agent: False for agent in _agent_ids} # Initialize truncateds for consistency

        # --- Store the experience for the initial timestep (timestep_0 or timestep_1 depending on convention) ---
        timestep_counter += 1
        timestep_data: Dict[str, Any] = {}
        for agent in _agent_ids:
            agent_experience: Dict[str, Any] = {
                'observation': obs[agent],
                'action': None, # No action taken yet at timestep 0
                'reward': reward[agent],
                'done': done["__all__"],
                'truncated': truncateds[agent], # Add truncateds
                'info': info[agent]
            }
            if rl_module_dict[agent].is_stateful(): # Check rnn_use for the specific agent
                agent_experience['state'] = current_states.get(agent) # Use .get() in case agent not in current_states
            timestep_data[agent] = agent_experience
        episode_data[f'timestep_{timestep_counter}'] = timestep_data
        
        # Loop to generate the experience for subsequent timesteps.
        while not done["__all__"]:
            # Action dictionary for the current timestep.
            action: Dict[str, Any] = {agent: None for agent in _agent_ids}
            obs_batch: Dict[str, Any] = {agent: None for agent in _agent_ids}
            
            # Get the action for each agent in the timestep.
            for agent in _agent_ids:
                                
                if rl_module_dict[agent].is_stateful(): # Check rnn_use for the specific agent
                    
                    # Compute the next action from a batch (B=1) of observations.
                    obs_batch[agent] = torch.from_numpy(obs[agent]).float().view(1, 1, -1)
                    # current_states[agent] = torch.from_numpy(np.array(current_states[agent].values(), dtype=np.float32)).unsqueeze(0)  # add batch B=1 dimension
                    current_states[agent] = tree.map_structure(
                        lambda x: x.unsqueeze(0) if x.ndim == 2 else x, 
                        current_states[agent]
                    )
                    prev_action_tensor = torch.tensor([[prev_action[agent]]], dtype=torch.int64)
                    
                    # Compute action with RNN state
                    input_dict: Dict[str, Any] = {
                        "obs": obs_batch[agent],
                        "state_in": current_states[agent],
                        "prev_actions": prev_action_tensor,
                        "seq_lens": torch.tensor([1], dtype=torch.int32)
                    }
                    
                    with torch.no_grad():
                        model_outputs = rl_module_dict[agent].forward_inference(input_dict)
                    
                    # Extract the action distribution parameters from the output and dissolve batch dim.
                    action_dist_params = model_outputs["action_dist_inputs"][0].numpy()
                    
                    # We have continuous actions -> take the mean (max likelihood).
                    # greedy_action = np.clip(
                    #     action_dist_params[0:1],  # 0=mean, 1=log(stddev), [0:1]=use mean, but keep shape=(1,)
                    #     a_min=env_instance.action_space[agent].low[0],
                    #     a_max=env_instance.action_space[agent].high[0],
                    # )
                    
                    # For discrete actions, you should take the argmax over the logits:
                    greedy_action = np.argmax(action_dist_params)
                    action[agent] = greedy_action
                    # new_states[agent] = agent_state
                    
                    # Update current_states for the next iteration if RNN is used by any agent
                    current_states[agent] = model_outputs["state_out"]
                    prev_action[agent] = action[agent]
                    
                else:
                    # Compute the next action from a batch (B=1) of observations.
                    obs_batch[agent] = torch.from_numpy(obs[agent]).float().unsqueeze(0)
                    # Compute action without RNN state
                    model_outputs  = rl_module_dict[agent].forward_inference({"obs": obs_batch[agent]})
                    # Extract the action distribution parameters from the output and dissolve batch dim.
                    action_dist_params = model_outputs["action_dist_inputs"][0].numpy()
                    # We have continuous actions -> take the mean (max likelihood).
                    # greedy_action = np.clip(
                    #     action_dist_params[0:1],  # 0=mean, 1=log(stddev), [0:1]=use mean, but keep shape=(1,)
                    #     a_min=env_instance.action_space[agent].low[0],
                    #     a_max=env_instance.action_space[agent].high[0],
                    # )
                    # For discrete actions, you should take the argmax over the logits:
                    greedy_action = np.argmax(action_dist_params)
                    action[agent] = greedy_action

                # assert type(action[agent]) is float|int, f"Action must be a float or int, but got {type(action[agent])} for agent {agent} at timestep {timestep_counter}."
            # Execute the action in the environment.
            new_obs, new_reward, done, truncateds, new_info = env_instance.step(action)

            # Update the observation, reward, done, truncateds, and info for the next timestep.
            obs = new_obs
            info = new_info
            reward = new_reward
            
            # List the agents in the environment for the next timestep.
            # This handles cases where agents might leave the environment.
            _agent_ids = list(obs.keys())
            
            # --- Store the experience for the current timestep ---
            timestep_counter += 1
            timestep_data = {}
            for agent in _agent_ids:
                agent_experience = {
                    'observation': obs[agent],
                    'action': action.get(agent), # Action taken to reach this state
                    'reward': reward[agent],
                    'done': done["__all__"],
                    'truncated': truncateds["__all__"],
                    'info': info[agent]
                }
                if rl_module_dict[agent].is_stateful(): # Check rnn_use for the specific agent
                    # Store the state *after* computing the action for this timestep
                    agent_experience['state'] = current_states.get(agent) # Use .get()
                timestep_data[agent] = agent_experience
            episode_data[f'timestep_{timestep_counter}'] = timestep_data

        # Add the completed episode data to the experiment data
        experiment_data['experiment'][f'episode_{episode_idx}'] = episode_data

    return experiment_data



"""
Specification for agent reward functions
=========================================
This module defines the ``RewardSpec`` class, which is used to specify the configuration of reward 
functions for agents in reinforcement learning environments.

It ensures that the reward function is properly defined and adheres to the expected interface.
"""
from typing import Dict, Any, Optional, Type

from eprllib.Agents.Rewards.BaseReward import BaseReward
from eprllib import logger

class RewardSpec:
    """
    RewardSpec is the base class for an reward specification to safe configuration of the object.
    """
    reward_fn: Optional[Type[BaseReward]] = None
    reward_fn_config: Dict[str, Any] = {}
        
    def __init__(
        self,
        reward_fn: Optional[Type[BaseReward]] = None,
        reward_fn_config: Dict[str, Any] = {},
        ):
        """
        Construction method.
        
        Args:
            reward_fn (BaseReward): The reward funtion take the arguments EnvObject (the GymEnv class) and the infos 
            dictionary. As a return, gives a float number as reward. See eprllib.Agents.Rewards for examples.
            reward_fn_config (Dict[str, Any]): The configuration of the reward function.
            
        Raises:
            NotImplementedError: If the reward_fn is NotImplemented.
        """
        self.reward_fn = reward_fn
        self.reward_fn_config = reward_fn_config
    
    def __getitem__(self, key:str):
        return getattr(self, key)

    def __setitem__(self, key:str, value:Any):
        valid_keys = self.__dict__.keys()
        if key not in valid_keys:
            msg = f"RewardSpec: Invalid key: {key}."
            logger.error(msg)
            raise KeyError(msg)
        setattr(self, key, value)
    
    def build(self) -> Dict[str, Any]:
        """
        This method is used to build the RewardSpec object.
        """
        if self.reward_fn == None:
            msg = "RewardSpec: No reward function provided."
            logger.error(msg)
            raise NotImplementedError(msg)
        
        return vars(self)
    


"""
Ramdom weather for episode
===========================

Given a path for a folder with epw files, this function update the epw file randomly in
each episode during the training process.
"""

from typing import Dict, Any
from eprllib.Episodes.BaseEpisode import BaseEpisode
from eprllib.Utils.episode_fn_utils import get_random_weather
from eprllib.Utils.annotations import override
from eprllib import logger

class RandomWeatherEpisode(BaseEpisode):
    """
    This class contains the methods to configure the episode in EnergyPlus with RLlib.
    
    Methods:
        get_episode_config: return the env_config with a updated epw path from the epw_files_folder_path.
    """
    def __init__(
        self,
        episode_fn_config: Dict[str,Any] = {}
        ):
        super().__init__(episode_fn_config)
        
        # check that 'epw_files_folder_path' exist in the episode_fn_config
        if 'epw_files_folder_path' not in self.episode_fn_config:
            msg = "The 'epw_files_folder_path' must be defined in the episode_fn_config."
            logger.error(msg)
            raise ValueError(msg)

    @override(BaseEpisode)
    def get_episode_config(self, env_config: Dict[str,Any]) -> Dict[str,Any]:
        """
        This method update the 'epw_path' property of the env_config Dict and returns it
        with the change appied.

        Returns:
            Dict: The episode configuration.
        """
        env_config['epw_path'] = get_random_weather(self.episode_fn_config['epw_files_folder_path'])
        return env_config
    


"""
Episode Functions
=================

This module contains the methods necesary to modify during runtime the environment for 
each episode. This allow to program different training steps like requiered by
curriculum learning.
"""



"""
Default Agents Connector
=========================

This module defines the default connector class that allows the combination of agents' observations 
to provide a flexible configuration of the communication between agents. Built-in hierarchical 
(only two levels), fully-shared, centralized, and independent configurations are provided.
"""
from typing import Dict, Any
from gymnasium.spaces import Box

from eprllib.Connectors.BaseConnector import BaseConnector
from eprllib.Utils.annotations import override
from eprllib.Utils.connector_utils import (
    set_variables_in_obs,
    set_internal_variables_in_obs,
    set_meters_in_obs,
    set_simulation_parameters_in_obs,
    set_zone_simulation_parameters_in_obs,
    set_prediction_variables_in_obs,
    set_other_obs_in_obs,
    set_actuators_in_obs,
    set_user_occupation_forecast_in_obs
    )
from eprllib import logger

class DefaultConnector(BaseConnector):
    
    @override(BaseConnector)
    def setup(self) -> None:
        """
        This method can be overridden in subclasses to perform setup tasks.
        """
        self.env_config: Dict[str, Any] = {}
    
    @override(BaseConnector)
    def get_agent_obs_dim(
        self,
        agent: str
        ) -> Box:
        """
        Get the agent observation dimension.
        
        Args:
            agent (str): Agent identifier.
        
        Returns:
            gym.spaces.Space: Agent observation dimension.
            
        Raises:
            NotImplementedError: If the method is not implemented in the child class.
        """
        obs_space_len: int = 0
        self.obs_indexed[agent] = {}
        
        self.obs_indexed[agent], obs_space_len = set_variables_in_obs(self.env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_internal_variables_in_obs(self.env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_meters_in_obs(self.env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_simulation_parameters_in_obs(self.env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_zone_simulation_parameters_in_obs(self.env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_prediction_variables_in_obs(self.env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_other_obs_in_obs(self.env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_actuators_in_obs(self.env_config, agent, self.obs_indexed[agent], obs_space_len)
        self.obs_indexed[agent], obs_space_len = set_user_occupation_forecast_in_obs(self.env_config, agent, self.obs_indexed[agent], obs_space_len)
                
        assert obs_space_len > 0, "The observation space length must be greater than 0."
        assert len(self.obs_indexed[agent]) == obs_space_len, f"The observation space length must be equal to the number of indexed observations. Obs indexed:{len(self.obs_indexed[agent])} != Obs space len:{obs_space_len}."
        # obs_space_len += 1
        logger.debug(f"DefaultConnector: Observation space length for agent {agent}: {obs_space_len}")
        
        return Box(float("-inf"), float("inf"), (obs_space_len, ))
    
    @override(BaseConnector)
    def get_agent_obs_indexed(
        self,
        agent: str
    ) -> Dict[str, int]:
        """
        Get a dictionary of the agent observation parameters and their respective index in the observation array.
        
        Args:
            agent (str): Agent identifier.
            
        Returns:
            Dict[str, int]: Dictionary of the agent observation parameters and their respective index in the observation array.
        
        Raises:
            NotImplementedError: If the method is not implemented in the child class.
        """
        if self.obs_indexed == {}:
            self.get_agent_obs_dim(agent)
        return self.obs_indexed[agent]
    